import rich.prompt
from rich.text import TextType

from exegol.utils.ExeLog import console, ConsoleLock


class ExegolRich:

    # These methods cannot be run from another Thread

    @staticmethod
    async def Ask(prompt: str, **kwargs) -> str:
        """Quick function to format rich Prompt and options on every exegol interaction"""
        async with ConsoleLock:
            return rich.prompt.Prompt.ask("[bold blue][?][/bold blue] " + prompt, console=console, **kwargs)

    @staticmethod
    async def Confirm(question: str, default: bool) -> bool:
        """Quick function to format rich Confirmation and options on every exegol interaction"""
        default_text = "[bright_magenta][Y/n][/bright_magenta]" if default else "[bright_magenta]\\[y/N][/bright_magenta]"
        formatted_question = f"[bold blue][?][/bold blue] {question} {default_text}"
        async with ConsoleLock:
            return rich.prompt.Confirm.ask(
                formatted_question,
                show_choices=False,
                show_default=False,
                console=console,
                default=default)

    @staticmethod
    async def Acknowledge(message: str) -> None:
        """Quick function to format rich Confirmation and options on every exegol interaction"""
        formatted_question = f"[bold blue][>][/bold blue] {message} [bright_magenta][Press ENTER to acknowledge][/bright_magenta]"
        async with ConsoleLock:
            try:
                rich.prompt.Prompt.ask(
                    formatted_question,
                    show_choices=False,
                    show_default=False,
                    console=console)
            except ValueError:
                raise EOFError



import binascii
import logging
import os
from asyncio import gather
from typing import Union, List, Tuple, Optional, cast, Sequence, Type

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.EnvInfo import EnvInfo
from exegol.config.UserConfig import UserConfig
from exegol.console import ConsoleFormat
from exegol.console.ConsoleFormat import boolFormatter
from exegol.console.ExegolPrompt import ExegolRich
from exegol.console.ExegolStatus import ExegolStatus
from exegol.console.TUI import ExegolTUI
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.console.cli.actions.GenericParameters import ContainerCreation
from exegol.exceptions.ExegolExceptions import ObjectNotFound, CancelOperation
from exegol.manager.LicenseManager import LicenseManager
from exegol.manager.TaskManager import TaskManager
from exegol.manager.UpdateManager import UpdateManager
from exegol.model.ExegolContainer import ExegolContainer
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.ExegolModules import ExegolModules
from exegol.model.SelectableInterface import SelectableInterface
from exegol.utils.DockerUtils import DockerUtils
from exegol.utils.ExeLog import logger, ExeLog
from exegol.utils.SessionHandler import SessionHandler


class ExegolManager:
    """Contains the main procedures of all actions available in Exegol"""

    # Cache data
    __container: Union[Optional[ExegolContainer], List[ExegolContainer]] = None
    __image: Union[Optional[ExegolImage], List[ExegolImage]] = None

    # Runtime default configuration
    __interactive_mode = False

    @classmethod
    async def info(cls) -> None:
        """Print a list of available images and containers on the current host"""
        if logger.isEnabledFor(ExeLog.VERBOSE):
            logger.verbose("Listing user configurations")
            ExegolTUI.printTable(UserConfig().get_configs(), title="[not italic]:brain: [/not italic][gold3][g]User configurations[/g][/gold3]")
        if logger.isEnabledFor(ExeLog.ADVANCED):
            logger.verbose("Listing git repositories")
            ExegolTUI.printTable(await UpdateManager.listGitStatus(), title="[not italic]:octopus: [/not italic][gold3][g]Project modules[/g][/gold3]")
        if bool(ParametersManager().containertag):
            # If the user have supplied a container name, show container config
            container = await cls.__loadOrCreateContainer(ParametersManager().containertag, must_exist=True)
            if container is not None:
                assert type(container) is ExegolContainer
                await ExegolTUI.printContainerRecap(container)
        else:
            # Without any parameter, show all images and containers info
            # Fetch data
            TaskManager.add_task(
                DockerUtils().listImages(include_version_tag=False, include_custom=True),
                TaskManager.TaskId.ImageList)
            TaskManager.add_task(
                DockerUtils().listContainers(),
                TaskManager.TaskId.ContainerList)
            images, containers = await TaskManager.gather(TaskManager.TaskId.ImageList, TaskManager.TaskId.ContainerList)
            # List and print images
            color = ConsoleFormat.getArchColor(ParametersManager().arch)
            logger.verbose(f"Listing local and remote Exegol images (filtering for architecture [{color}]{ParametersManager().arch}[/{color}])")
            ExegolTUI.printTable(images)
            # List and print containers
            logger.verbose("Listing local Exegol containers")
            logger.raw(f"[bold blue][*][/bold blue] Number of Exegol containers: {len(containers)}{os.linesep}",
                       markup=True)
            ExegolTUI.printTable(containers)

    @classmethod
    async def start(cls) -> None:
        """Create and/or start an exegol container to finally spawn an interactive shell"""
        # Check if the first positional parameter have been supplied
        cls.__interactive_mode = not bool(ParametersManager().containertag)
        if not cls.__interactive_mode:
            logger.info("Skipping interactive mode (arguments supplied)")
        container = await cls.__loadOrCreateContainer()
        assert container is not None and type(container) is ExegolContainer
        if not container.isNew():
            # Check and warn user if some parameters don't apply to the current session
            cls.__checkUselessParameters()
        await gather(
            container.start(),
            TaskManager.wait_for_all())
        await container.spawnShell()

    @classmethod
    async def upgrade(cls) -> None:
        """Upgrade exegol to the latest version"""
        if not SessionHandler().pro_feature_access():
            logger.critical("Exegol upgrade is only available for Pro or Enterprise users.")
        container = await cls.__loadOrCreateContainer(multiple=True, must_exist=True, filters=[ExegolContainer.Filters.OUTDATED])
        assert container is not None and type(container) is list
        for c in container:
            try:
                previous_image = c.image
                await cls.__backupAndUpgrade(c)

                # If the image used is deprecated, it must be deleted after the removal of its container
                if previous_image.isLocked() and UserConfig().auto_remove_images:
                    await DockerUtils().removeImage(previous_image, upgrade_mode=True, silent_error=True)
            except CancelOperation:
                logger.error(f"Something unexpected happened during the [green]{c.name}[/green] container upgrade process.")

    @classmethod
    async def exec(cls) -> None:
        """Create and/or start an exegol container to execute a specific command.
        The execution can be seen in console output or be relayed in the background as a daemon."""
        if ParametersManager().tmp:
            container = await cls.__createTmpContainer(ParametersManager().selector)
            if not ParametersManager().daemon:
                await container.exec(command=ParametersManager().exec, as_daemon=False, is_tmp=True)
                await container.stop(timeout=2)
            else:
                # Command is passed at container creation in __createTmpContainer()
                logger.success(f"Command executed as entrypoint of the container {container.getDisplayName()}")
        else:
            container = cast(ExegolContainer, await cls.__loadOrCreateContainer(override_container=ParametersManager().selector))
            await container.exec(command=ParametersManager().exec, as_daemon=ParametersManager().daemon)

    @classmethod
    async def stop(cls) -> None:
        """Stop an exegol container"""
        logger.info("Stopping container(s)")
        container = await cls.__loadOrCreateContainer(multiple=True, must_exist=True, filters=[ExegolContainer.Filters.STARTED])
        assert container is not None and type(container) is list
        for c in container:
            await c.stop(timeout=5)

    @classmethod
    async def restart(cls) -> None:
        """Stop and start an exegol container"""
        container = cast(ExegolContainer, await cls.__loadOrCreateContainer(must_exist=True))
        if container:
            await container.stop(timeout=5)
            await container.start()
            logger.success(f"Container [green]{container.name}[/green] successfully restarted!")
            await container.spawnShell()

    @classmethod
    async def install(cls) -> None:
        """Pull or build a docker exegol image"""
        try:
            if not await ExegolModules().isExegolResourcesReady():
                raise CancelOperation
        except CancelOperation:
            # Error during installation, skipping operation
            logger.warning("Exegol resources have not been downloaded, the feature cannot be enabled")
        await UpdateManager.updateImage(install_mode=True)

    @classmethod
    async def build(cls) -> None:
        """Build a docker exegol image"""
        await UpdateManager.buildAndLoad(load_after_build=False)

    @classmethod
    async def update(cls) -> None:
        """Update python wrapper (git installation required) and Pull a docker exegol image"""
        if ParametersManager().offline_mode:
            logger.critical("Exegol cannot be updated without Internet access. Skipping.")
        if not ParametersManager().skip_git:
            await UpdateManager.updateWrapper()
            await UpdateManager.updateResources()
        if not ParametersManager().skip_images:
            await UpdateManager.updateImage()

    @classmethod
    async def uninstall(cls) -> None:
        """Remove an exegol image"""
        logger.info("Uninstalling image")
        # Set log level to verbose in order to show every image installed including the outdated.
        if not logger.isEnabledFor(ExeLog.VERBOSE):
            logger.setLevel(ExeLog.VERBOSE)
        images = await cls.__loadOrInstallImage(multiple=True, filters=[ExegolImage.Filters.INSTALLED])
        assert type(images) is list
        if len(images) == 0:
            return
        all_name = ", ".join([x.getName() for x in images])
        if not ParametersManager().force_mode and not await ExegolRich.Confirm(
                f"Are you sure you want to [red]permanently remove[/red] the following images? [orange3][ {all_name} ][/orange3]",
                default=False):
            logger.error("Aborting operation.")
            return
        for img in images:
            await DockerUtils().removeImage(img)

    @classmethod
    async def remove(cls) -> None:
        """Remove an exegol container"""
        logger.info("Removing container(s)")
        containers = await cls.__loadOrCreateContainer(multiple=True, must_exist=True)
        assert type(containers) is list
        if len(containers) == 0:
            logger.error("No containers were selected. Exiting.")
            return
        all_name = ", ".join([x.name for x in containers])
        if not ParametersManager().force_mode and not await ExegolRich.Confirm(
                f"Are you sure you want to [red]permanently remove[/red] the following containers? [orange3][ {all_name} ][/orange3]",
                default=False):
            logger.error("Aborting operation.")
            return
        for c in containers:
            # Check if containers have backups
            backup_history = c.getExistingBackupContainers()
            if len(backup_history) > 0:
                all_backup_names = ', '.join([x[1] for x in backup_history])
                if ParametersManager().force_mode:
                    logger.info(f"The container [green]{c.name}[/green] has [green]{len(backup_history)}[/green] backup containers that will be removed too [orange3][ {all_backup_names} ][/orange3]")
                elif not await ExegolRich.Confirm(f"The container [green]{c.name}[/green] has [green]{len(backup_history)}[/green] backup containers [orange3][ {all_backup_names} ][/orange3], do you want to remove them too?", default=True):
                    logger.critical("Cannot remove container without also removing its backups. Aborting.")

            await c.remove(backup_history=[x[0] for x in backup_history])
            # If the image used is deprecated, it must be deleted after the removal of its container
            if c.image.isLocked() and UserConfig().auto_remove_images:
                await DockerUtils().removeImage(c.image, upgrade_mode=True)

    @classmethod
    async def activate(cls) -> None:
        """Activate an exegol license"""
        if ParametersManager().revoke:
            await (await LicenseManager.get()).revoke_exegol()
        elif (await LicenseManager.get()).is_activated():
            logger.success("Exegol is activated")
        else:
            await (await LicenseManager.get()).activate_exegol(skip_prompt=True)

    @classmethod
    async def print_version(cls) -> None:
        """Show exegol version (and context configuration on debug mode)"""

        logger.raw(f"[bold blue][*][/bold blue] Exegol {SessionHandler().get_license_type_display()} is currently in version {await UpdateManager.display_current_version()}{os.linesep}",
                   level=logging.INFO, markup=True)
        await LicenseManager.get()
        logger.raw(
            f"[bold magenta][*][/bold magenta] More about Exegol at: [underline magenta]{ConstantConfig.landing}[/underline magenta]{os.linesep}",
            level=logging.INFO, markup=True)

        cls.print_sponsors()
        if 'a' in ConstantConfig.version:
            logger.empty_line()
            logger.warning("You are currently using an [red]Alpha[/red] version of Exegol, which may be unstable. "
                           "This version is a work in progress and bugs are expected.")
        elif 'b' in ConstantConfig.version:
            logger.empty_line()
            logger.warning("You are currently using a [orange3]Beta[/orange3] version of Exegol, which may be unstable.")

    @classmethod
    async def print_debug_banner(cls) -> None:
        """Print header debug info"""
        package_engine = "pip"
        if ConstantConfig.pipx_installed:
            package_engine = "pipx"
        elif ConstantConfig.uv_installed:
            package_engine = "uv"
        logger.debug(f"Pip installation: {boolFormatter(ConstantConfig.pip_installed)} [bright_black]({package_engine})[/bright_black]")
        logger.debug(f"Git source installation: {boolFormatter(ConstantConfig.git_source_installation)}")
        logger.debug(f"Host OS: {EnvInfo.getHostOs().value} [bright_black]({EnvInfo.getDockerEngine().value})[/bright_black]")
        logger.debug(f"Arch: {EnvInfo.arch}")
        if EnvInfo.arch != EnvInfo.raw_arch:
            logger.debug(f"Raw arch: {EnvInfo.raw_arch}")
        if EnvInfo.isWindowsHost():
            logger.debug(f"Windows release: {EnvInfo.getWindowsRelease()}")
            logger.debug(f"Python environment: {EnvInfo.current_platform}")
            logger.debug(f"Docker engine: {EnvInfo.getDockerEngine().value}")
        logger.debug(f"Docker desktop: {boolFormatter(EnvInfo.isDockerDesktop())}")
        logger.debug(f"Shell type: {EnvInfo.getShellType()}")
        if UserConfig().auto_check_updates:
            await UpdateManager.checkForWrapperUpdate()
        if await UpdateManager.isUpdateAvailable():
            logger.empty_line()
            update_message = f"An [green]Exegol[/green] update is [orange3]available[/orange3] ({await UpdateManager.display_current_version()} :arrow_right: {UpdateManager.display_latest_version()})"
            if ConstantConfig.git_source_installation:
                if await ExegolRich.Confirm(f"{update_message}, do you want to update ?", default=True):
                    await UpdateManager.updateWrapper()
            else:
                logger.info(update_message)
                update_command = None
                if ConstantConfig.pipx_installed:
                    update_command = "You can update your exegol wrapper with the command [green]pipx upgrade exegol[/green]"
                elif ConstantConfig.uv_installed:
                    update_command = "You can update your exegol wrapper with the command [green]uv tool upgrade exegol[/green]"
                elif ConstantConfig.pip_installed:
                    update_command = "If you have installed Exegol with pip, update with the command [green]pip3 install exegol --upgrade[/green]"
                else:
                    await ExegolRich.Acknowledge("Installation method not found (not among pip/pipx/uv/sources). You should update your wrapper manually.")
                if update_command:
                    await ExegolRich.Acknowledge(update_command)
        else:
            logger.empty_line(log_level=logging.DEBUG)

    @classmethod
    def print_sponsors(cls) -> None:
        """Show exegol sponsors"""
        # logger.success("""We thank [link=https://www.capgemini.com/fr-fr/carrieres/offres-emploi/][blue]Capgemini[/blue][/link] for supporting the project [bright_black](helping with dev)[/bright_black] :pray:""")
        # logger.success("""We thank [link=https://www.hackthebox.com/][green]HackTheBox[/green][/link] for sponsoring the [bright_black]multi-arch[/bright_black] support :green_heart:""")
        pass

    @classmethod
    async def __loadOrInstallImage(cls,
                                   override_image: Optional[str] = None,
                                   multiple: bool = False,
                                   show_custom: bool = False,
                                   filters: Optional[List[ExegolImage.Filters]] = None) -> Union[Optional[ExegolImage], List[ExegolImage]]:
        """Select / Load (and install) an ExegolImage
        When multiple is set to True, return a list of ExegolImage
        When the action supports multi selection, the parameter filters must be supplied
        When filters include ExegolImage.Filters.INSTALLED, return None if no image are installed
        Otherwise, always return an ExegolImage"""
        if cls.__image is not None:
            # Return cache
            return cls.__image
        must_exist = filters is not None and ExegolImage.Filters.INSTALLED in filters
        image_tag = override_image if override_image is not None else ParametersManager().imagetag
        image_tags = ParametersManager().multiimagetag
        image_selection: Union[Optional[ExegolImage], List[ExegolImage]] = None
        # While an image have not been selected
        while image_selection is None:
            try:
                if image_tag is None and (image_tags is None or len(image_tags) == 0):
                    image_list: List[ExegolImage] = await DockerUtils().listImages(include_custom=show_custom)
                    if filters is not None:
                        filters_sum = sum(filters)
                        image_list = [i for i in image_list if i.filter(filters_sum)]

                    if ParametersManager().select_all:
                        image_selection = image_list
                    else:
                        # Interactive (TUI) image selection
                        image_selection = cast(Union[Optional[ExegolImage], List[ExegolImage]],
                                               await cls.__interactiveSelection(ExegolImage, image_list, multiple))
                else:
                    # Select image by tag name (non-interactive)
                    if multiple:
                        image_selection = []
                        for image_tag in image_tags:
                            image_selection.append(await DockerUtils().getInstalledImage(image_tag))
                    else:
                        image_selection = await DockerUtils().getInstalledImage(image_tag)
            except ObjectNotFound:
                # ObjectNotFound is raised when the image_tag provided by the user does not match any existing image.
                if image_tag is not None:
                    logger.warning(f"The image named '{image_tag}' has not been found.")
                # If the user's selected image have not been found,
                # offer to build a local image with this name
                # (only if must_exist is not set)
                if not must_exist:
                    image_selection = await UpdateManager.updateImage(image_tag)
                # Allow the user to interactively select another installed image
                image_tag = None
            except IndexError:
                # IndexError is raised when no image are available (not applicable when multiple is set, return an empty array)
                # (raised from TUI interactive selection)
                if must_exist:
                    # If there is no image installed, return none
                    logger.error("Nothing to do.")
                    return [] if multiple else None
                elif image_tag is not None:
                    # If the user's selected image have not been found, offer the choice to build a local image at this name
                    # (only if must_exist is not set)
                    image_selection = await UpdateManager.updateImage(image_tag)
                    image_tag = None
                else:
                    logger.critical("No image are installed or available, check your internet connection and install an image with the command [green]exegol install[/green].")
            # Checks if an image has been selected
            if image_selection is None:
                # If not, retry the selection
                logger.error("No image has been selected.")
                continue

            # Check if every image are installed
            install_status, checked_images = await cls.__checkImageInstallationStatus(image_selection, multiple, must_exist)
            if not install_status:
                # If one of the image is not install where it supposed to, restart the selection
                # allowing him to interactively choose another image
                image_selection, image_tag = None, None
                continue

            cls.__image = cast(Union[Optional[ExegolImage], List[ExegolImage]], checked_images)
        return cls.__image

    @classmethod
    async def __checkImageInstallationStatus(cls,
                                             image_selection: Union[ExegolImage, List[ExegolImage]],
                                             multiple: bool = False,
                                             must_exist: bool = False
                                             ) -> Tuple[bool, Optional[Union[ExegolImage, ExegolContainer, List[ExegolImage], List[ExegolContainer]]]]:
        """Checks if the selected images are installed and ready for use.
        returns false if the images are supposed to be already installed."""
        # Checks if one or more images have been selected and unifies the format into a list.
        reverse_type = False
        check_img: List[ExegolImage]
        if type(image_selection) is ExegolImage:
            check_img = [image_selection]
            # Tag of the operation to reverse it before the return
            reverse_type = True
        elif type(image_selection) is list:
            check_img = image_selection
        else:
            check_img = []

        # Check if every image are installed
        for i in range(len(check_img)):
            if not check_img[i].isInstall():
                # Is must_exist is set, every image are supposed to be already installed
                if must_exist:
                    logger.error(f"The selected image '{check_img[i].getName()}' is not installed.")
                    # If one of the image is not install, return False to restart the selection
                    return False, None
                else:
                    # Check if the selected image is installed and install it
                    logger.warning("The selected image is not installed.")
                    # Download remote image
                    if await DockerUtils().downloadImage(check_img[i], install_mode=True):
                        # Select installed image
                        check_img[i] = await DockerUtils().getInstalledImage(check_img[i].getName(), check_img[i].getRepository())
                    else:
                        logger.error("This image cannot be installed.")
                        return False, None

        if reverse_type and not multiple:
            # Restoration of the original type
            return True, check_img[0]
        return True, check_img

    @classmethod
    async def __loadOrCreateContainer(cls,
                                      override_container: Optional[str] = None,
                                      multiple: bool = False,
                                      must_exist: bool = False,
                                      filters: Optional[List[ExegolContainer.Filters]] = None) -> Union[Optional[ExegolContainer], List[ExegolContainer]]:
        """Select one or more ExegolContainer
        Or create a new ExegolContainer if no one already exist (and must_exist is not set)
        When must_exist is set to True, return None if no container exist
        When multiple is set to True, return a list of ExegolContainer"""
        if cls.__container is not None:
            # Return cache
            return cls.__container
        container_tag: Optional[str] = override_container if override_container is not None else ParametersManager().containertag
        container_tags: Optional[List[str]] = None
        if ParametersManager().multicontainertag:
            container_tags = []
            for tag in ParametersManager().multicontainertag:
                # Prevent duplicate tag selection
                if tag not in container_tags:
                    container_tags.append(tag)
        try:
            if container_tag is None and (container_tags is None or len(container_tags) == 0):
                container_list: List[ExegolContainer] = await DockerUtils().listContainers()
                if filters is not None:
                    filters_sum = sum(filters)
                    container_list = [c for c in container_list if c.filter(filters_sum)]
                if ParametersManager().select_all:
                    # Select all container
                    cls.__container = container_list
                else:
                    # Interactive container selection
                    cls.__container = cast(Union[Optional[ExegolContainer], List[ExegolContainer]],
                                           await cls.__interactiveSelection(ExegolContainer, container_list, multiple))
            else:
                # Try to find the corresponding container
                if multiple:
                    cls.__container = []
                    assert container_tags is not None
                    # test each user tag
                    for container_tag in container_tags:
                        try:
                            cls.__container.append(DockerUtils().getContainer(container_tag))
                        except ObjectNotFound:
                            # on multi select, an object not found is not critical
                            if must_exist:
                                # If the selected tag doesn't match any container, print an alert and continue
                                logger.warning(f"The container named '{container_tag}' has not been found")
                            else:
                                # If there is a multi select without must_exist flag, raise an error
                                # because multi container creation is not supported
                                raise NotImplementedError
                else:
                    assert container_tag is not None
                    cls.__container = DockerUtils().getContainer(container_tag)
        except (ObjectNotFound, IndexError):
            # ObjectNotFound is raised when the container_tag provided by the user does not match any existing container.
            # IndexError is raise when no container exist (raised from TUI interactive selection)
            # Create container
            if must_exist:
                if container_tag is not None:
                    logger.warning(f"Container '{container_tag}' has not been found")
                return [] if multiple else None
            logger.info(f"Creating new container '{container_tag}'")
            return await cls.__createContainer(container_tag)
        assert cls.__container is not None
        return cast(Union[Optional[ExegolContainer], List[ExegolContainer]], cls.__container)

    @classmethod
    async def __interactiveSelection(cls,
                                     object_type: Type[Union[ExegolImage, ExegolContainer]],
                                     object_list: Sequence[SelectableInterface],
                                     multiple: bool = False) -> \
            Union[Optional[ExegolImage], Optional[ExegolContainer], Sequence[ExegolImage], Sequence[ExegolContainer]]:
        """Interactive object selection process, depending on object_type.
        object_type can be ExegolImage or ExegolContainer."""
        user_selection: Union[SelectableInterface, Sequence[SelectableInterface], Sequence[str], str]
        if multiple:
            user_selection = await ExegolTUI.multipleSelectFromTable(object_list, object_type=object_type)
        else:
            user_selection = await ExegolTUI.selectFromTable(object_list, object_type=object_type, allow_none=object_type is ExegolContainer)
            # Check if the user has chosen an existing object
            if type(user_selection) is str:
                # Otherwise, create a new object with the supplied name
                if object_type is ExegolContainer:
                    user_selection = await cls.__createContainer(user_selection)
        return cast(Union[ExegolImage, ExegolContainer, List[ExegolImage], List[ExegolContainer]], user_selection)

    @classmethod
    async def __createContainer(cls, name: Optional[str]) -> ExegolContainer:
        """Create an ExegolContainer"""
        if name is None:
            name = await ExegolRich.Ask("Enter new container name", default="default")
        logger.verbose("Configuring new exegol container")
        # Create exegol config
        image: Optional[ExegolImage] = cast(ExegolImage, await cls.__loadOrInstallImage(show_custom=True))
        assert image is not None  # load or install return an image
        if name is None:
            name = await ExegolRich.Ask("Enter new container name", default="default")
        model = await ExegolContainerTemplate.newContainer(name, image, hostname=ParametersManager().hostname)

        # Recap
        await ExegolTUI.printContainerRecap(model)
        if cls.__interactive_mode:
            if not model.image.isUpToDate() and "Unknown" not in model.image.getStatus() and \
                    await ExegolRich.Confirm("Do you want to [green]update[/green] the selected image?", False):
                image = await UpdateManager.updateImage(model.image.getName())
                if image is not None:
                    model.image = image
                    await ExegolTUI.printContainerRecap(model)
            command_options = []
            while not await ExegolRich.Confirm("Is the container configuration [green]correct[/green]?", default=True):
                command_options = await model.config.interactiveConfig(model.name)
                await ExegolTUI.printContainerRecap(model)
            logger.info(f"Command line of the configuration: "
                        f"[green]exegol start {model.name} {model.image.getName()} {' '.join(command_options)}[/green]")
            logger.info("To use exegol [orange3]without interaction[/orange3], "
                        "read CLI options with [green]exegol start -h[/green]")

        container = DockerUtils().createContainer(model)
        await container.postCreateSetup()
        return container

    @classmethod
    async def __createTmpContainer(cls, image_name: Optional[str] = None) -> ExegolContainer:
        """Create a temporary ExegolContainer with custom entrypoint"""
        logger.verbose("Configuring new exegol container")
        name = f"tmp-{binascii.b2a_hex(os.urandom(4)).decode('ascii')}"
        # Create exegol config
        image: ExegolImage = cast(ExegolImage, await cls.__loadOrInstallImage(override_image=image_name))
        model = await ExegolContainerTemplate.newContainer(name, image, hostname=ParametersManager().hostname)
        # When container exec a command as a daemon, the execution must be set on the container's entrypoint
        if ParametersManager().daemon:
            # Using formatShellCommand to support zsh aliases
            exec_payload, str_cmd = ExegolContainer.formatShellCommand(ParametersManager().exec, entrypoint_mode=True)
            model.config.entrypointRunCmd()
            model.config.addEnv("CMD", str_cmd)
            model.config.addEnv("DISABLE_AUTO_UPDATE", "true")
        # Workspace must be disabled for temporary container because host directory is never deleted
        model.config.disableDefaultWorkspace()

        # Mount entrypoint as a volume (because in tmp mode the container is created with run instead of create method)
        model.config.addVolume(ConstantConfig.entrypoint_context_path_obj, "/.exegol/entrypoint.sh", must_exist=True, read_only=True)

        container = DockerUtils().createContainer(model, temporary=True)
        await container.postCreateSetup(is_temporary=True)
        return container

    @classmethod
    def __checkUselessParameters(cls) -> None:
        """Checks if the container creation parameters have not been filled in when the container already existed"""
        # Get defaults parameters
        creation_parameters = ContainerCreation([]).__dict__
        # Get parameters from user input
        user_inputs = ParametersManager().parameters.__dict__
        detected = []
        for param in creation_parameters.keys():
            # Skip parameters useful in a start context
            if param in ('containertag',):
                continue
            # For each parameter, check if it's not None and different from the default
            current_option = creation_parameters.get(param)
            if user_inputs.get(param) is not None and current_option is not None and \
                    user_inputs.get(param) != current_option.kwargs.get('default'):
                # If the supplied parameter is positional, getting his printed name
                name = current_option.kwargs.get('metavar')
                if not name:
                    # if not, using the args name
                    detected.append(' / '.join(current_option.args))
                else:
                    detected.append(name)
        if len(detected) > 0:
            logger.warning(f"These parameters ({', '.join(detected)}) have been entered although the container already "
                           f"exists, they will not be taken into account.")

    @classmethod
    async def __backupAndUpgrade(cls, c: ExegolContainer) -> None:
        logger.empty_line()

        current_image_tag = c.image.getName().split('-')[0]
        if ParametersManager().image_tag is None or ParametersManager().image_tag == current_image_tag:

            # Check update conditions
            if not c.image.isLocked():
                if "Unknown" in c.image.getStatus():
                    await c.image.autoLoad()
                if c.image.isUpToDate():
                    logger.error(f"Container [green]{c.name}[/green] is already using the latest version of [blue]{current_image_tag}[/blue]. No need to upgrade, skipping.")
                    return

            # Tips to upgrade from free image
            if current_image_tag == "free":
                logger.info(f"[orange3][Tips][/orange3] you can use the [green]--image IMAGE[/green] option to upgrade your container to a {SessionHandler().get_license_type_display()} image (e.g., 'full'').")
                logger.empty_line()

            new_image: ExegolImage = await DockerUtils().getInstalledImage(current_image_tag)
            logger.info(f"Upgrading container [green]{c.name}[/green] using [green]{new_image.getName()}[/green] image")
        else:
            # Upgrade to a different image tag
            new_image = await DockerUtils().getInstalledImage(ParametersManager().image_tag)
            logger.info(f"Upgrading container [green]{c.name}[/green], your container will migrate from [blue]{current_image_tag}[/blue] to the [blue]{new_image.getName()}[/blue] image")

        skipping_msg = ""
        if not new_image.isUpToDate():
            skipping_msg = f"Run [green]exegol update {new_image.getName()}[/green] to install the new version [green]{new_image.getLatestVersion()}[/green] of your image first."
            logger.warning(f"You're upgrading to an outdated version of the [blue]{new_image.getName()}[/blue] image ([orange3]{new_image.getImageVersion()}[/orange3] :arrow_right: [green]{new_image.getLatestVersion()}[/green])")
            if not ParametersManager().force_mode and not await ExegolRich.Confirm(f"Are you sure you want to upgrade your container [green]{c.name}[/green] to an outdated image?", default=False):
                if await ExegolRich.Confirm(f"Do you want to update your [green]{new_image.getName()}[/green] image now?", default=False):
                    image_update = await UpdateManager.updateImage(new_image.getName())
                    if image_update is None:
                        logger.error(f"An error occurred during image update. Skipping upgrade of container [green]{c.name}[/green].")
                        return
                    new_image = image_update
                else:
                    logger.info(f"Skipping upgrade of container [green]{c.name}[/green]. {skipping_msg}")
                    return

        # Check if the new image is the same as the container's current image
        if c.image.getLocalId() == new_image.getLocalId():
            logger.error(f"Cannot upgrade [green]{c.name}[/green] because it's already using the latest local [blue]{new_image.getName()}[/blue] image (version [orange3]{new_image.getImageVersion()}[/orange3]), skipping.")
            if not new_image.isUpToDate():
                logger.info(skipping_msg)
            return

        # Start container and run pre-backup checks
        if not c.isRunning():
            await c.start()
        async with ExegolStatus(f"Running pre-backup checks", spinner_style="blue"):
            # Test if exh can be backup
            backup_directory_exist = await c.exec(f"[ -d {ExegolContainer.BACKUP_DIRECTORY} ]", as_daemon=False, quiet=True, show_output=False) == 0
            if not backup_directory_exist:
                exh_backup_supported = await c.exec("exegol-history version", as_daemon=False, quiet=True, show_output=False) == 0

        if backup_directory_exist:
            logger.error(f"The directory {ExegolContainer.BACKUP_DIRECTORY} already exists in the container [green]{c.name}[/green]. "
                         f"It needs to be removed manually before trying to upgrade this container.")
            return

        remove_container = ParametersManager().no_backup or (
                    not ParametersManager().force_mode and
                    not await ExegolRich.Confirm("Do you want to [green]keep[/green] your old container as a backup?", default=True))

        backup_items = [
            "Your [green]my-resources[/green] customization" if c.config.isMyResourcesEnable() else "",
            "The container [green]/workspace[/green] directory",
            "Your [green]bash/zsh[/green] command history",
            "Your [green]exegol-history[/green] database" if exh_backup_supported else "",
            "Your [green]TriliumNext[/green] notes",
            "The following files: /etc/hosts /etc/resolv.conf /opt/tools/Exegol-history/profile.sh",
            "The following configurations: [green]Proxychains[/green]"
        ]
        backup_text = '\n    - '.join([i for i in backup_items if i])
        details = f"""You are about to upgrade your container and transfer:
    - {backup_text}
"""
        # TODO improve upgrade with
        #  Config of: Responder?
        #  DB of Responder, neo4j, postgres, nxc?, firefox, hashcat potfile, john?

        logger.warning(details)
        if (not ParametersManager().force_mode and
                not await ExegolRich.Confirm(f"The list above will be [orange3]{'kept' if remove_container else 'transferred'} "
                                             f"to the new container[/orange3], [red]nothing more{', without backup' if remove_container else ''}[/red]! "
                                             f"Do you want to proceed with the upgrade of [green]{c.name}[/green]?", default=False)):
            logger.critical("Aborting operation.")

        logger.warning("Please don't cancel this operation while it's running! You might loose some data!")

        # Start container data Backup
        await c.backup(backup_exh=exh_backup_supported)
        logger.success(f"Container [green]{c.name}[/green] data has been backed up.")

        # Get previous backups that still exist
        backup_history = c.getExistingBackupContainers()

        if remove_container:
            # Remove the container without removing the workspace
            await c.remove(container_only=True)
        else:
            await c.stop()
            # Renaming old container
            c.rename_as_old()
            # Add and update backup history references
            backup_history.append((c.getFullId(), ''))

        # Updating previous backup containers references
        c.config.setBackupHistory(','.join([x[0] for x in backup_history]) if len(backup_history) > 0 else None)

        # Update container's exegol image to the new image target
        c.image = new_image

        # Create a new container from template
        container = DockerUtils().createContainer(c)
        await container.postCreateSetup()

        # Restore data on new container
        if await container.restore():
            logger.success(f"Container [green]{c.name}[/green] successfully upgraded to the [green]{c.image.getLatestVersionName().replace('-', ' ')}[/green] image!")
        else:
            logger.warning("The container was upgraded, but errors occurred during data restoration. Consult the previous messages to recover the missing data in your new container")
        logger.info(f"You can now open a shell in your new container with [green]exegol start {c.name}[/green]")



import hashlib
import platform
import uuid

from exegol.utils.ExeLog import logger
from exegol.utils.LocalDatastore import LocalDatastore


class MUID:

    @classmethod
    def __compute(cls, rid: str) -> str:
        uname = platform.uname()
        data = [rid,
                uname.processor,
                uname.machine,
                uname.version,
                uname.release,
                uname.system,
                uname.node,
                "/".join(platform.architecture())]
        return hashlib.sha3_256('+'.join(data).encode("utf-16")).hexdigest()

    @classmethod
    def get_current_muid(cls) -> str:
        logger.debug("Computing MUID")
        rid, mid = LocalDatastore().get_machine_id()
        if rid is not None:
            current_id = cls.__compute(rid)
            if mid == current_id:
                return current_id
        rid = str(uuid.uuid4())
        current_id = cls.__compute(rid)
        LocalDatastore().update_mid(rid, current_id)
        return current_id



import io
import tarfile
from typing import Optional

from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger


class ImageScriptSync:

    @staticmethod
    def getCurrentStartVersion() -> str:
        """Find the current version of the spawn.sh script."""
        with open(ConstantConfig.spawn_context_path_obj, 'r') as file:
            for line in file.readlines():
                if line.startswith('# Spawn Version:'):
                    return line.split(':')[-1].strip()
        logger.critical(f"The spawn.sh version cannot be found, check your exegol setup! {ConstantConfig.spawn_context_path_obj}")
        raise RuntimeError

    @staticmethod
    def getImageSyncTarData(include_entrypoint: bool = False, include_spawn: bool = False) -> Optional[bytes]:
        """The purpose of this class is to generate and overwrite scripts like the entrypoint or spawn.sh inside exegol containers
        to integrate the latest features, whatever the version of the image."""

        # Create tar file
        stream = io.BytesIO()
        with tarfile.open(fileobj=stream, mode='w|') as entry_tar:

            # Load entrypoint data
            if include_entrypoint:
                entrypoint_script_path = ConstantConfig.entrypoint_context_path_obj
                logger.debug(f"Entrypoint script path: {str(entrypoint_script_path)}")
                if not entrypoint_script_path.is_file():
                    logger.critical("Unable to find the entrypoint script! Your Exegol installation is probably broken...")
                    return None
                with open(entrypoint_script_path, 'rb') as f:
                    raw = f.read()
                    data = io.BytesIO(initial_bytes=raw)

                # Import file to tar object
                info = tarfile.TarInfo(name="/.exegol/entrypoint.sh")
                info.size = len(raw)
                info.mode = 0o500
                entry_tar.addfile(info, fileobj=data)

            # Load start data
            if include_spawn:
                spawn_script_path = ConstantConfig.spawn_context_path_obj
                logger.debug(f"Spawn script path: {str(spawn_script_path)}")
                if not spawn_script_path.is_file():
                    logger.error("Unable to find the spawn script! Your Exegol installation is probably broken...")
                    return None
                with open(spawn_script_path, 'rb') as f:
                    raw = f.read()
                    data = io.BytesIO(initial_bytes=raw)

                # Import file to tar object
                info = tarfile.TarInfo(name="/.exegol/spawn.sh")
                info.size = len(raw)
                info.mode = 0o500
                entry_tar.addfile(info, fileobj=data)
        return stream.getvalue()



# Exceptions specific to the successful operation of exegol
class ObjectNotFound(Exception):
    """Custom exception when a specific container do not exist"""
    pass


class ProtocolNotSupported(Exception):
    """Custom exception when a specific network protocol is not supported"""
    pass


class CancelOperation(Exception):
    """Custom exception when an error occurred and the operation must be canceled ou skipped"""
    pass


class LicenseRevocation(Exception):
    """Custom exception when the license has been revoked"""
    pass


class LicenseToleration(Exception):
    pass


class InteractiveError(Exception):
    """Custom exception when an error occurred. In the interactive mode this error can be retry be the user, in CLI mode this is a critical error"""
    pass



import asyncio
import http
import logging

try:
    import docker
    import git
    import requests
    import urllib3
    import supabase
    import httpx
    import postgrest
    import contextlib

    from exegol.utils.ExeLog import logger, ExeLog, console
    from exegol.utils.DockerUtils import DockerUtils
    from exegol.console.cli.ParametersManager import ParametersManager
    from exegol.console.cli.actions.ExegolParameters import Command
    from exegol.manager.ExegolManager import ExegolManager
    from exegol.manager.TaskManager import TaskManager
    from exegol.utils.SessionHandler import SessionHandler
except ModuleNotFoundError as e:
    print("Mandatory dependencies are missing:", e)
    print("Please install them with python3 -m pip install --upgrade -r requirements.txt")
    exit(1)
except ImportError as e:
    print("An error occurred while loading the dependencies!")
    print()
    if "git executable" in e.msg:
        print("Git is missing in your PATH, it must be installed locally on your computer.")
        print()
    print("Details:")
    print(e)
    exit(1)
except KeyboardInterrupt:
    exit(1)


class ExegolController:
    """Main controller of exegol"""

    # Get action selected by user
    # (ParametersManager must be loaded from ExegolController first to load every Command subclass)
    __action: Command = ParametersManager().getCurrentAction()

    @classmethod
    async def call_action(cls) -> int:
        """Dynamically retrieve the main function corresponding to the action selected by the user
        and execute it on the main thread"""
        try:
            await ExegolManager.print_version()
            DockerUtils()  # Init dockerutils
            await ExegolManager.print_debug_banner()
            # Check for missing parameters
            missing_params = cls.__action.check_parameters()
            if len(missing_params) == 0:
                # Fetch main operation function
                main_action = cls.__action()
                return_code = 0
                if main_action is not None:
                    TaskManager.add_task(
                        SessionHandler().reload_session(),
                        TaskManager.TaskId.LoadLicense)
                    # Execute main function
                    await main_action()
                await TaskManager.wait_for_all()
                return return_code
            else:
                # TODO review required parameters
                logger.error(f"These parameters are mandatory but missing: {','.join(missing_params)}")
                return 1
        except SystemExit as err:
            logger.empty_line()
            logger.info("Exiting...")
            await TaskManager.wait_for_all(exit_mode=True)
            return 1 if err.code is None else int(err.code)
        except (KeyboardInterrupt, asyncio.CancelledError, EOFError):
            logger.empty_line()
            logger.info("Exiting...")
            await TaskManager.wait_for_all(exit_mode=True)
            return 1


def print_exception_banner() -> None:
    logger.error("It seems that something unexpected happened ...")
    logger.error("To draw our attention to the problem and allow us to fix it, you can share your error with us "
                 "(by [orange3]copying and pasting[/orange3] it with this syntax: ``` <error> ```) "
                 "by creating a GitHub issue at this address: https://github.com/ThePorgs/Exegol/issues")
    logger.success("Thank you for your collaboration!")


def main() -> int:
    """Exegol main console entrypoint"""
    try:
        # Set logger verbosity depending on user input
        ExeLog.setVerbosity(ParametersManager().verbosity, ParametersManager().quiet)
        # Start Main controller & Executing action selected by user CLI
        return asyncio.run(ExegolController.call_action())
    except (KeyboardInterrupt, asyncio.CancelledError, EOFError):
        return 2
    except git.exc.GitCommandError as git_error:
        print_exception_banner()
        # Printing git stderr as raw to avoid any Rich parsing error
        logger.debug("Full git output:")
        logger.raw(git_error, level=logging.DEBUG)
        logger.empty_line()
        error = git_error.stderr.strip().split(": ")[-1].strip("'")
        logger.error("Git error received:")
        # Printing git error as raw to avoid any Rich parsing error
        logger.raw(error, level=logging.ERROR)
        logger.empty_line()
        logger.error(f"A critical error occurred while running this git command: {' '.join(git_error.command)}")
    except SystemExit as e:
        if e.code is not None:
            return int(e.code)
    except Exception:
        print_exception_banner()
        console.print_exception(show_locals=True, suppress=[docker, requests, git, urllib3, http, httpx, postgrest, contextlib, supabase, asyncio])
    return 1



from typing import Optional, Union, Dict

from docker.models.images import Image


class MetaImages:
    """Meta model to store and organise multi-arch images from registry API"""

    def __init__(self, digest: str, image_name: str, tag_name: str, images_size: Dict[str, int]) -> None:
        """Create a MetaImage object to handle multi-arch docker registry images in a single point"""
        self.tag_name: str = tag_name
        self.image: str = image_name
        self.images_size: Dict[str, int] = images_size  # Arch : size
        self.images_size_left = self.images_size.copy()
        self.meta_id: str = digest
        self.version: str = '-'.join(self.tag_name.split('-')[1:])
        self.is_latest: bool = not bool(self.version)  # Current image is latest if no version have been found from tag name

    def __str__(self) -> str:
        return f"{self.tag_name} ({self.version}) [{self.meta_id}] {self.images_size.keys()}"

    def __repr__(self) -> str:
        return self.__str__()



from typing import cast, Union

from rich.progress import Progress, Task, TaskID, ProgressColumn

from exegol.utils.ExeLog import console, ConsoleLock


class ExegolProgress(Progress):
    """Addition of a practical function to Rich Progress"""

    def __init__(self, *columns: Union[str, ProgressColumn], **kwargs) -> None:
        super().__init__(*columns, console=console, **kwargs)

    def getTask(self, task_id: TaskID) -> Task:
        """Return a specific task from task_id without error"""
        task = self._tasks.get(task_id)
        if task is None:
            # If task doesn't exist, raise IndexError exception
            raise IndexError
        return cast(Task, task)

    def __enter__(self) -> "ExegolProgress":
        super(ExegolProgress, self).__enter__()
        return self

    async def __aenter__(self) -> "ExegolProgress":
        await ConsoleLock.acquire()
        try:
            super(ExegolProgress, self).__enter__()
            return self
        except Exception as e:
            ConsoleLock.release()
            raise e

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            self.__exit__(exc_type, exc_val, exc_tb)
        finally:
            ConsoleLock.release()



import argparse
from logging import CRITICAL
from typing import Optional, List, Union, Dict, cast

import argcomplete

from exegol.config.ConstantConfig import ConstantConfig
from exegol.console.cli.actions.Command import Command, Option
from exegol.utils.ExeLog import logger


class ExegolArgParse(argparse.ArgumentParser):
    """Overloading of the main parsing (argparse.ArgumentParser) class"""

    # Using Exelog to print built-in parser message
    def _print_message(self, message: str, file=None) -> None:
        if message:
            logger.raw(message, level=CRITICAL, rich_parsing=True)


class Parser:
    """Custom Exegol CLI Parser. Main controller of argument building and parsing."""

    __description = f"""This Python script is a wrapper for Exegol. It can be used to easily manage Exegol on your machine.

[bold magenta]Exegol documentation:[/bold magenta] [underline magenta]{ConstantConfig.documentation}[/underline magenta]"""
    __formatter_class = argparse.RawTextHelpFormatter

    def __init__(self, actions: List[Command]):
        """Custom parser creation"""
        # Defines every actions available
        self.__actions: List[Command] = actions
        # Create & init root parser
        self.__root_parser: ExegolArgParse
        self.__init_parser()
        # Configure root options
        # (WARNING: these parameters are duplicate with every sub-parse, cannot use positional args here)
        self.__set_options(self.__root_parser, Command())  # Add global arguments from Command to the root parser
        # Create & fill sub-parser
        self.subParser = self.__root_parser.add_subparsers(help="Description of the actions")
        self.__set_action_parser()

    def __init_parser(self) -> None:
        """Root parser creation"""

        self.__root_parser = ExegolArgParse(
            description=self.__description,
            epilog=Command().formatEpilog(),
            formatter_class=self.__formatter_class,
        )

    def __set_action_parser(self) -> None:
        """Create sub-parser for each action and configure it"""
        self.__root_parser._positionals.title = "[green]Required arguments[/green]"
        for action in self.__actions:
            # Each action has a dedicated sub-parser with different options
            # the 'help' description of the current action is retrieved
            # from the comment of the corresponding action class
            if action.__doc__ is None:
                action.__doc__ = "Unknown action"
            sub_parser = self.subParser.add_parser(action.name, help=action.__doc__,
                                                   description=action.__doc__ + f"""\n
[bold magenta]Exegol documentation:[/bold magenta] [underline magenta]{ConstantConfig.documentation}/wrapper/cli/{action.name}[/underline magenta]""",
                                                   epilog=action.formatEpilog(),
                                                   formatter_class=self.__formatter_class)
            sub_parser.set_defaults(action=action)
            self.__set_options(sub_parser, target=action)

    def __set_options(self, sub_parser: argparse.ArgumentParser, target: Optional[Command] = None) -> None:
        """Add different groups and parameters/options in the current sub_parser"""
        global_set = False  # Only one group can be global at the time
        # Load actions to be processed (default: every action from cls)
        actions_list = [target] if target else self.__actions
        for action in actions_list:
            # On each action, fetch every group to be processed
            for argument_group in action.groupArgs:
                group_parser: argparse._ActionsContainer
                if argument_group.is_global and not global_set:
                    # If the current group is global (ex: 'Optional arguments'),
                    # overwriting parser main group before adding custom parameters
                    global_set = True  # The setup application should be run only once
                    sub_parser._optionals.title = argument_group.title  # Overwriting default argparse title
                    group_parser = sub_parser  # Subparser is directly used to add arguments
                else:
                    # In every other case, a dedicated group is created in the parser
                    group_parser = sub_parser.add_argument_group(argument_group.title,
                                                                 description=argument_group.description)
                # once the group is created in the parser, the arguments can be added to it
                option: Dict[str, Union[Option, bool]]
                for option in argument_group.options:
                    # Retrieve Option object from the Dict
                    assert type(option["arg"]) is Option
                    argument = cast(Option, option["arg"])
                    # Pop is required here to removed unknown parameter from the action object before argparse
                    completer = argument.kwargs.pop("completer", None)
                    try:
                        arg = group_parser.add_argument(*argument.args, **argument.kwargs)
                    except argparse.ArgumentError:
                        continue
                    # Add argument with its config to the parser
                    if completer is not None:
                        arg.completer = completer  # type: ignore

    def run_parser(self) -> argparse.Namespace:
        """Execute argparse to retrieve user options from argv"""
        argcomplete.autocomplete(self.__root_parser)
        return self.__root_parser.parse_args()

    def print_help(self) -> None:
        """Force argparse to display the help message"""
        self.__root_parser.print_help()



class SelectableInterface:
    """Generic class used to select objects in the user TUI"""

    def getKey(self) -> str:
        """Universal unique key getter"""
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        """Generic '==' operator overriding matching object key"""
        if type(other) is str:
            return other == self.getKey()
        elif isinstance(other, SelectableInterface):
            return other.getKey() == self.getKey()
        raise NotImplementedError



import os
import sqlite3
import sys
import threading
from enum import Enum
from typing import Tuple, Optional, Union, cast

from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger
from exegol.utils.FsUtils import get_user_id
from exegol.utils.MetaSingleton import MetaSingleton


class LocalDatastore(metaclass=MetaSingleton):
    __DB_PATH = ConstantConfig.exegol_config_path / ".datastore"

    class Key(Enum):
        EULA = 0
        SESSION_CERT = 1
        TOKEN = 2
        SESSION = 3

    def __init__(self) -> None:

        sqlite3.threadsafety = 3
        self.__db_lock = threading.Lock()

        self.__is_init = self.__DB_PATH.is_file()
        try:
            self.__db = sqlite3.connect(self.__DB_PATH,
                                        check_same_thread=False,
                                        autocommit=False)  # type: ignore[call-overload]
        except TypeError:
            # autocommit available from python 3.12, before that using isolation_level
            self.__db = sqlite3.connect(self.__DB_PATH,
                                        check_same_thread=False,
                                        isolation_level="IMMEDIATE")

        # For a new installation, set the owner of the DB to the current user
        if not self.__is_init and sys.platform == "linux" and os.getuid() == 0:
            user_uid, user_gid = get_user_id()
            os.chown(self.__DB_PATH, user_uid, user_gid)

        self.__apply_schema()

    def __apply_schema(self) -> None:
        with self.__db_lock, self.__db:
            try:
                self.__db.execute("CREATE TABLE IF NOT EXISTS machine (rid TEXT NOT NULL, mid TEXT NOT NULL)")
                self.__db.execute("CREATE TABLE IF NOT EXISTS kv (key SMALLINT NOT NULL UNIQUE, value TEXT NOT NULL)")
                self.__db.commit()
            except sqlite3.Error as e:
                self.__db.rollback()
                raise e

    # EULA SECTION

    def is_eula_accepted(self) -> bool:
        return self.get(self.Key.EULA) == "1"

    def update_eula(self, value: bool) -> None:
        self.set(self.Key.EULA, "1" if value else "0")

    # LICENSE SECTION

    def get_license(self) -> Tuple[Optional[str], Optional[str]]:
        token = self.get(LocalDatastore.Key.TOKEN)
        if type(token) is bytes:
            token = token.decode("utf-8")
        session = self.get(LocalDatastore.Key.SESSION)
        if type(session) is bytes:
            session = session.decode("utf-8")
        return cast(Optional[str], session), cast(Optional[str], token)

    def deactivate_license(self) -> None:
        with self.__db_lock, self.__db:
            logger.debug("DB Deactivating license")
            try:
                self.__db.execute("DELETE FROM kv WHERE key = ?", (self.Key.TOKEN.value,))
                self.__db.execute("DELETE FROM kv WHERE key = ?", (self.Key.SESSION.value,))
                self.__db.commit()
            except sqlite3.Error as e:
                logger.error("DB error during: Deactivating license")
                self.__db.rollback()
                raise e

    # MACHINE SECTION
    def get_machine_id(self) -> Tuple[Optional[str], Optional[str]]:
        cursor = self.__db.cursor()
        result = cursor.execute("SELECT * FROM machine").fetchone()
        cursor.close()
        if result is None:
            return None, None
        return result

    def update_mid(self, rid: str, mid: str) -> None:
        with self.__db_lock, self.__db:
            logger.debug("DB Updating DB MID")
            try:
                self.__db.execute("DELETE FROM machine")
                self.__db.execute("INSERT INTO machine (rid, mid) VALUES (?, ?)", (rid, mid,))
                self.__db.commit()
            except sqlite3.Error as e:
                logger.error("DB error during: Updating DB MID")
                self.__db.rollback()
                raise e

    # KV SECTION
    def get(self, key: Key) -> Optional[Union[str, bytes]]:
        with self.__db_lock, self.__db:
            result = self.__db.execute("SELECT value FROM kv WHERE key = ?", (key.value,)).fetchone()
        if result is None:
            return None
        return result[0]

    def set(self, key: Key, value: Optional[Union[str, bytes]]) -> None:
        with self.__db_lock, self.__db:
            logger.debug(f"DB Updating KV {key.name}")
            try:
                if value is not None:
                    self.__db.execute("REPLACE INTO kv (key, value) VALUES (?, ?)", (key.value, value,))
                else:
                    self.__db.execute("DELETE FROM kv WHERE key = ?", (key.value,))
                self.__db.commit()
            except sqlite3.Error as e:
                logger.error(f"DB error during: Updating DB KV {key.name}")
                self.__db.rollback()
                raise e
        return None



from typing import List, Optional

from argcomplete.completers import EnvironCompleter, DirectoriesCompleter, FilesCompleter

from exegol.config.UserConfig import UserConfig
from exegol.console.cli.ExegolCompleter import ContainerCompleter, ImageCompleter, VoidCompleter, DesktopConfigCompleter
from exegol.console.cli.SyntaxFormat import SyntaxFormat
from exegol.console.cli.actions.Command import Option, GroupArg
from exegol.utils.NetworkUtils import NetworkUtils


class ContainerSelector:
    """Generic parameter class for container selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create container selector arguments
        self.containertag: Optional[Option] = Option("containertag",
                                                     metavar="CONTAINER",
                                                     nargs='?',
                                                     action="store",
                                                     help="Tag used to target an Exegol container",
                                                     completer=ContainerCompleter)

        # Create group parameter for container selection
        groupArgs.append(GroupArg({"arg": self.containertag, "required": False},
                                  title="[blue]Container selection options[/blue]"))


class ContainerMultiSelector:
    """Generic parameter class for container multi selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create container selector arguments
        self.multicontainertag = Option("multicontainertag",
                                        metavar="CONTAINER",
                                        nargs='*',
                                        action="store",
                                        help="Tag used to target one or more Exegol containers",
                                        completer=ContainerCompleter)
        self.select_all = Option("--all",
                                 dest="select_all",
                                 action="store_true",
                                 default=False,
                                 help="Select every Exegol containers available")

        # Create group parameter for container multi selection
        groupArgs.append(GroupArg({"arg": self.multicontainertag, "required": False},
                                  {"arg": self.select_all, "required": False},
                                  title="[blue]Containers selection options[/blue]"))


class ContainerStart:
    """Generic parameter class for container selection.
    This generic class is used by start, restart and exec actions"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create options on container start
        self.envs = Option("-e", "--env",
                           action="append",
                           default=[],
                           dest="envs",
                           help="And an environment variable on Exegol (format: --env KEY=value). The variables "
                                "configured during the creation of the container will be persistent in all shells. "
                                "If the container already exists, the variable will be present only in the current shell",
                           completer=EnvironCompleter)

        self.capabilities = Option("--cap",
                                   dest="capabilities",
                                   metavar='CAPABILITY',  # Do not display available choices
                                   action="append",
                                   default=[],
                                   choices={"NET_ADMIN", "NET_BROADCAST", "SYS_MODULE", "SYS_PTRACE", "SYS_RAWIO",
                                            "SYS_ADMIN", "LINUX_IMMUTABLE", "MAC_ADMIN", "SYSLOG", "ALL"},
                                   help="[orange3](dangerous)[/orange3] Capabilities allow to add [orange3]specific[/orange3] privileges to the container "
                                        "(e.g. need to mount volumes, perform low-level operations on the network, etc).")

        # Create group parameter for container options at start
        groupArgs.append(GroupArg({"arg": self.envs, "required": False},
                                  {"arg": self.capabilities, "required": False},
                                  title=f"[blue]Container creation or start options[/blue]"))


class ContainerSpawnShell(ContainerStart):
    """Generic parameter class to spawn a shell on an exegol container.
    This generic class is used by start and restart"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Spawn container shell arguments
        self.shell = Option("-s", "--shell",
                            dest="shell",
                            action="store",
                            choices=UserConfig.start_shell_options,
                            default=UserConfig().default_start_shell,
                            help=f"Select a shell environment to launch at startup (Default: [blue]{UserConfig().default_start_shell}[/blue])")

        self.log = Option("-l", "--log",
                          dest="log",
                          action="store_true",
                          default=False,
                          help="Enable shell logging (commands and outputs) on exegol to /workspace/logs/ (default: [red]Disabled[/red])")
        self.log_method = Option("--log-method",
                                 dest="log_method",
                                 action="store",
                                 choices=UserConfig.shell_logging_method_options,
                                 default=UserConfig().shell_logging_method,
                                 help=f"Select a shell logging method used to record the session (default: [blue]{UserConfig().shell_logging_method}[/blue])")
        self.log_compress = Option("--log-compress",
                                   dest="log_compress",
                                   action="store_true",
                                   default=False,
                                   help=f"Enable or disable the automatic compression of log files at the end of the session (default: {'[green]Enabled[/green]' if UserConfig().shell_logging_compress else '[red]Disabled[/red]'})")

        # Group dedicated to shell logging feature
        groupArgs.append(GroupArg({"arg": self.log, "required": False},
                                  {"arg": self.log_method, "required": False},
                                  {"arg": self.log_compress, "required": False},
                                  title="[blue]Container shell logging options[/blue]"))

        ContainerStart.__init__(self, groupArgs)

        # Create group parameter for container selection
        groupArgs.append(GroupArg({"arg": self.shell, "required": False},
                                  title="[bold cyan]Start[/bold cyan] [blue]specific options[/blue]"))


class ImageSelector:
    """Generic parameter class for image selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create image selector arguments
        self.imagetag: Optional[Option] = Option("imagetag",
                                                 metavar="IMAGE",
                                                 nargs='?',
                                                 action="store",
                                                 help="Tag used to target an Exegol image",
                                                 completer=ImageCompleter)

        # Create group parameter for image selection
        groupArgs.append(GroupArg({"arg": self.imagetag, "required": False},
                                  title="[blue]Image selection options[/blue]"))


class ImageMultiSelector:
    """Generic parameter class for image multi selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create image multi selector arguments
        self.multiimagetag = Option("multiimagetag",
                                    metavar="IMAGE",
                                    nargs='*',
                                    action="store",
                                    help="Tag used to target one or more Exegol images",
                                    completer=ImageCompleter)
        self.select_all = Option("--all",
                                 dest="select_all",
                                 action="store_true",
                                 default=False,
                                 help="Select every Exegol images available")

        # Create group parameter for image multi selection
        groupArgs.append(GroupArg({"arg": self.multiimagetag, "required": False},
                                  {"arg": self.select_all, "required": False},
                                  title="[blue]Images selection options[/blue]"))


class ContainerCreation(ContainerSelector, ImageSelector):
    """Generic parameter class for container creation"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Init parents : ContainerStart > ContainerSelector
        ContainerSelector.__init__(self, groupArgs)
        ImageSelector.__init__(self, groupArgs)

        self.X11 = Option("--disable-X11",
                          action="store_false",
                          default=True,
                          dest="X11",
                          help="Disable X11 sharing to run GUI-based applications (default: [green]Enabled[/green])")
        self.my_resources = Option("--disable-my-resources",
                                   action="store_false",
                                   default=True,
                                   dest="my_resources",
                                   help=f"Disable the mount of the my-resources (/opt/my-resources) from the host ({UserConfig().my_resources_path}) (default: [green]Enabled[/green])")
        self.exegol_resources = Option("--disable-exegol-resources",
                                       action="store_false",
                                       default=True,
                                       dest="exegol_resources",
                                       help=f"Disable the mount of the exegol resources (/opt/resources) from the host ({UserConfig().exegol_resources_path}) (default: [green]Enabled[/green])")
        self.network = Option("--network",
                              dest="network",
                              action="store",
                              default=None,
                              choices=NetworkUtils.get_options(),
                              help=f"Select the type of network to which the container will be attached (default: [blue]{UserConfig().network_default_mode}[/blue])")
        self.share_timezone = Option("--disable-shared-timezones",
                                     action="store_false",
                                     default=True,
                                     dest="share_timezone",
                                     help="Disable the sharing of the host's time and timezone configuration with exegol (default: [green]Enabled[/green])")
        self.mount_current_dir = Option("-cwd", "--cwd-mount",
                                        dest="mount_current_dir",
                                        action="store_true",
                                        default=False,
                                        help="This option is a shortcut to set the /workspace folder to the user's current working directory")
        self.workspace_path = Option("-w", "--workspace",
                                     dest="workspace_path",
                                     action="store",
                                     help="The specified host folder will be linked to the /workspace folder in the container",
                                     completer=DirectoriesCompleter())
        self.update_fs_perms = Option("-fs", "--update-fs",
                                      action="store_true",
                                      default=False,
                                      dest="update_fs_perms",
                                      help=f"Modifies the permissions of folders and sub-folders shared in your workspace to access the files created within the container using your host user account. "
                                           f"(default: {'[green]Enabled[/green]' if UserConfig().auto_update_workspace_fs else '[red]Disabled[/red]'})")
        self.volumes = Option("-V", "--volume",
                              action="append",
                              default=[],
                              dest="volumes",
                              help=f"Share a new volume between host and exegol (format: --volume {SyntaxFormat.volume})")
        self.ports = Option("-p", "--port",
                            action="append",
                            default=[],
                            dest="ports",
                            help=f"Share a network port between host and exegol (format: --port {SyntaxFormat.port_sharing}). This configuration will disable the default host network.",
                            completer=VoidCompleter)
        self.hostname = Option("--hostname",
                               dest="hostname",
                               default=None,
                               action="store",
                               help="Set a custom hostname to the exegol container (default: exegol-<name>)",
                               completer=VoidCompleter)
        self.privileged = Option("--privileged",
                                 dest="privileged",
                                 action="store_true",
                                 default=False,
                                 help="[orange3](dangerous)[/orange3] Give [red]ALL[/red] admin privileges to the container when it is created "
                                      "(if the need is specifically identified, consider adding capabilities instead).")
        self.devices = Option("-d", "--device",
                              dest="devices",
                              default=[],
                              action="append",
                              help="Add host [default not bold]device(s)[/default not bold] at the container creation (example: -d /dev/ttyACM0 -d /dev/bus/usb/)")

        self.comment = Option("--comment",
                              dest="comment",
                              action="store",
                              help="The specified comment will be added to the container info",
                              completer=VoidCompleter)

        groupArgs.append(GroupArg({"arg": self.workspace_path, "required": False},
                                  {"arg": self.mount_current_dir, "required": False},
                                  {"arg": self.update_fs_perms, "required": False},
                                  {"arg": self.volumes, "required": False},
                                  {"arg": self.ports, "required": False},
                                  {"arg": self.hostname, "required": False},
                                  {"arg": self.privileged, "required": False},
                                  {"arg": self.devices, "required": False},
                                  {"arg": self.X11, "required": False},
                                  {"arg": self.my_resources, "required": False},
                                  {"arg": self.exegol_resources, "required": False},
                                  {"arg": self.network, "required": False},
                                  {"arg": self.share_timezone, "required": False},
                                  {"arg": self.comment, "required": False},
                                  title="[blue]Container creation options[/blue]"))

        self.vpn = Option("--vpn",
                          dest="vpn",
                          default=None,
                          action="store",
                          help="Setup an OpenVPN (.ovpn) or WireGuard (.conf) connection at the container creation (example: --vpn /home/user/vpn/client.ovpn)",
                          completer=FilesCompleter(["ovpn"], directories=True))
        self.vpn_auth = Option("--vpn-auth",
                               dest="vpn_auth",
                               default=None,
                               action="store",
                               help="Enter the credentials with a file (first line: username, second line: password) to establish the OpenVPN connection automatically (example: --vpn-auth /home/user/vpn/auth.txt)")

        groupArgs.append(GroupArg({"arg": self.vpn, "required": False},
                                  {"arg": self.vpn_auth, "required": False},
                                  title="[blue]Container creation VPN options[/blue]"))

        self.desktop = Option("--desktop",
                              dest="desktop",
                              action="store_true",
                              default=False,
                              help=f"Enable or disable the Exegol desktop feature (default: {'[green]Enabled[/green]' if UserConfig().desktop_default_enable else '[red]Disabled[/red]'})")
        self.desktop_config = Option("--desktop-config",
                                     dest="desktop_config",
                                     default="",
                                     action="store",
                                     help=f"Configure your exegol desktop ([blue]{'[/blue] or [blue]'.join(UserConfig.desktop_available_proto)}[/blue]) and its exposure "
                                          f"(format: {SyntaxFormat.desktop_config}) "
                                          f"(default: [blue]{UserConfig().desktop_default_proto}[/blue]:[blue]{'127.0.0.1' if UserConfig().desktop_default_localhost else '0.0.0.0'}[/blue]:[blue]<random>[/blue])",
                                     completer=DesktopConfigCompleter)
        groupArgs.append(GroupArg({"arg": self.desktop, "required": False},
                                  {"arg": self.desktop_config, "required": False},
                                  title="[blue]Container creation Desktop options[/blue]"))



from datetime import datetime, timezone
from typing import Optional, cast

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from exegol.exceptions.ExegolExceptions import CancelOperation
from exegol.utils.ExeLog import logger
from exegol.utils.LocalDatastore import LocalDatastore
from exegol.utils.MetaSingleton import MetaSingleton
from exegol.utils.SupabaseUtils import SupabaseUtils


class KeyHandler(metaclass=MetaSingleton):
    __CA = b"""-----BEGIN CERTIFICATE-----
MIICFDCCAXWgAwIBAgIUb1vW91MI9RmY3ckWPu/2zPjiG74wCgYIKoZIzj0EAwIw
IzEhMB8GA1UEAwwYRVhFQ09SUCBST09UIExpY2Vuc2VzIENBMCAXDTI1MDYwMTE1
MjIyMFoYDzIwNzUwNTIwMTUyMjE5WjAjMSEwHwYDVQQDDBhFWEVDT1JQIFJPT1Qg
TGljZW5zZXMgQ0EwgZswEAYHKoZIzj0CAQYFK4EEACMDgYYABAEBp2+vdMrKc4Ty
3SRGJ3/Dxo1q6c90a8Ii37GCAvpSHkQlOra5IZ9YDkw4jhWR/gXVrCiJsYcT1ITt
T7C+Rw/yNQDvEJlWrbQuEFosE7pd0JhVLL1Gb2HlfvFN3SR5+Qre5/unQzMgRhQ/
Ao3ThWVwVeGMOYQwChRDrRFvN0GQh1hWIqNCMEAwDwYDVR0TAQH/BAUwAwEB/zAd
BgNVHQ4EFgQU0zajN86zCCZXjj4jjcdgJ4O3xfAwDgYDVR0PAQH/BAQDAgGGMAoG
CCqGSM49BAMCA4GMADCBiAJCANcis/iuCO8MU337t65f9kpshUCi99Yhw3j/UrbF
2PxDnlB6q1EIpzkh10PQXwm0YW5rg7lMoCGSmm4L7cR3TZDBAkIBhiXNB+bJkDDE
bK214R8IIiJQyepYRvXdWJDflwoWJVhpBIZAHFdKkcw2oJOeRT+eIkmHgxvRHfp7
nvWygzUoHuc=
-----END CERTIFICATE-----"""

    __DEFAULT_PUB_CERT = b"""-----BEGIN CERTIFICATE-----
MIICLTCCAY6gAwIBAgIUc0Ve4qKUxu6aPZ3f5TsRxM/d3UwwCgYIKoZIzj0EAwIw
IzEhMB8GA1UEAwwYRVhFQ09SUCBST09UIExpY2Vuc2VzIENBMB4XDTI1MDYwMTE1
MjIyMFoXDTI4MDUzMTE1MjEzOFowQjEQMA4GA1UECgwHRVhFQ09SUDEQMA4GA1UE
CwwHV1JBUFBFUjEcMBoGA1UEAwwTbGljZW5zZXMuZXhlZ29sLmNvbTBZMBMGByqG
SM49AgEGCCqGSM49AwEHA0IABI+3RkjgH1t1H2gwaRSvoL3oh66/pm4SNcK15/WD
3eG6crZN1+KvATvAJELRiUJeSFqE26AAi7bfstFCYKQ41d+jgYAwfjAMBgNVHRMB
Af8EAjAAMB8GA1UdIwQYMBaAFNM2ozfOswgmV44+I43HYCeDt8XwMB4GA1UdEQQX
MBWCE2xpY2Vuc2VzLmV4ZWdvbC5jb20wHQYDVR0OBBYEFA7OvalfKxJoyZRk5W9c
ULYlyhnwMA4GA1UdDwEB/wQEAwIGwDAKBggqhkjOPQQDAgOBjAAwgYgCQgC4HNlY
ai3+0mXhY0QeYoYFaE2NqyKlpR5DobPw1o1UnQvcOIuHPXGQEIBDUT5Z0175YX74
1+5Jhaz6ILkRLtwxQQJCAPKgauhl0LMsZ9F/QxnnxPi1uJLHLhjf3hj9vjjknaXU
QSipTBpex0X2AoeZ4nLuVLlo8DXYziBk5YFBRKCjQ7xA
-----END CERTIFICATE-----"""

    def __init__(self) -> None:
        # Load current cert from cache
        raw_cert = LocalDatastore().get(LocalDatastore.Key.SESSION_CERT)
        if raw_cert is None:
            # Default cache setup
            raw_cert = self.__DEFAULT_PUB_CERT
            LocalDatastore().set(LocalDatastore.Key.SESSION_CERT, raw_cert)
        if type(raw_cert) is str:
            raw_cert = raw_cert.encode("utf-8")
        assert type(raw_cert) is bytes
        self.__cert: Optional[x509.Certificate] = x509.load_pem_x509_certificate(raw_cert)
        self.__ca_cert = x509.load_pem_x509_certificate(self.__CA)
        if self.__ca_cert.serial_number != 635746069563818933303724671059320780600564128702:
            logger.critical("Wrapper corrupted. Please do not edit your wrapper files. Contact support if needed.")
        if self.__ca_cert.not_valid_after_utc < datetime.now(timezone.utc):
            logger.critical("Your wrapper version is too old. Please update your wrapper.")

        # Cert attributes
        self.__subject: str = ''
        self.__serial_number: str = ''

        if self.__cert is not None and self.__verify_cert(self.__cert):
            self.__update_cert_meta()
        else:
            self.__cert = None

    def __update_cert_meta(self) -> None:
        if self.__cert is None:
            return
        # Cert attributes
        cn = self.__cert.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0].value
        self.__subject = cn.decode("utf-8") if type(cn) is bytes else cast(str, cn)
        self.__serial_number = "{:x}".format(self.__cert.serial_number).upper()

    def __verify_cert(self, cert: x509.Certificate) -> bool:
        subject = cert.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[0].value
        if type(subject) is bytes:
            subject = subject.decode("utf-8")
        assert type(subject) is str

        builder = x509.verification.PolicyBuilder()
        builder = builder.store(x509.verification.Store([self.__ca_cert]))
        verifier = builder.build_server_verifier(x509.DNSName(subject))
        try:
            verifier.verify(cert, [])
        except x509.verification.VerificationError as e:
            logger.debug(f"Certificate verification failed: {e}")
            return False
        return True

    def getSubject(self) -> str:
        return f"urn:exegol:license_srv:{self.__serial_number}"

    def getKey(self) -> bytes:
        if self.__cert is None:
            raise CancelOperation("Certificate not loaded")
        return self.__cert.public_key().public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)

    async def refresh_certificate(self, force_refresh: bool = False) -> bool:
        """
        Reload certificate from license server. Return true if the cert have been updated
        :param force_refresh:
        :return:
        """
        cert_changed = False
        # Default refresh only if needed
        if self.__cert is None or force_refresh:
            new_cert_raw = await SupabaseUtils.get_cert()
            new_cert = x509.load_pem_x509_certificate(new_cert_raw)
            if self.__verify_cert(new_cert):
                LocalDatastore().set(LocalDatastore.Key.SESSION_CERT, new_cert_raw)
                cert_changed = self.__cert is None or self.__cert.serial_number != new_cert.serial_number
                self.__cert = new_cert
                self.__update_cert_meta()
            else:
                logger.critical("Your exegol version is no longer compatible. Please update your wrapper or contact support.")
        return cert_changed



from enum import Enum
from typing import Optional
from typing_extensions import NotRequired, TypedDict


# Types
class LicenseType(Enum):
    Community = 0
    Professional = 1
    Enterprise = 2


# Form data
class EnrollmentForm(TypedDict):
    machine_id: str
    machine_name: str
    machine_os: str
    license_id: str
    revoke_previous_machine: NotRequired[bool]


# Response data
class LicensesEnumeration(TypedDict):
    type: str
    valid_until: str
    organization: str
    enrolled_on: str
    last_seen: str


class TokenRotate(TypedDict):
    next_token: str


class LicenseSession(TypedDict):
    session: Optional[str]


class LicenseEnrollment(TypedDict):
    next_token: str
    session: str


class SessionData(TypedDict):
    machine_id: str
    license: str
    license_id: str
    license_owner: str
    username: str
    expiration_date: str



import re
from pathlib import Path
from typing import Optional, Dict, cast, Tuple, Sequence, List

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.DataCache import DataCache
from exegol.config.UserConfig import UserConfig
from exegol.console.ExegolPrompt import ExegolRich
from exegol.console.ExegolStatus import ExegolStatus
from exegol.console.TUI import ExegolTUI
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import ObjectNotFound, CancelOperation
from exegol.model.ExegolImage import ExegolImage
from exegol.model.ExegolModules import ExegolModules
from exegol.utils.DockerUtils import DockerUtils
from exegol.utils.ExeLog import logger, ExeLog
from exegol.utils.GitUtils import GitUtils
from exegol.utils.WebRegistryUtils import WebRegistryUtils


class UpdateManager:
    """Procedure class for updating the exegol tool and docker images"""

    @classmethod
    async def updateImage(cls, tag: Optional[str] = None, install_mode: bool = False) -> Optional[ExegolImage]:
        """User procedure to build/pull docker image"""
        # List Images
        image_args = ParametersManager().imagetag
        # Select image
        if image_args is not None and tag is None:
            tag = image_args
        if tag is None:
            all_images: List[ExegolImage] = await DockerUtils().listImages()
            # Filter for updatable images
            if install_mode:
                available_images = [i for i in all_images if not i.isLocked() and not i.isLocal()]
            else:
                available_images = [i for i in all_images if i.isInstall() and not i.isLocal() and not i.isUpToDate() and not i.isLocked()]
                if len(available_images) == 0:
                    logger.success("All images already installed are up to date!")
                    return None
            try:
                # Interactive selection
                selected_image = await ExegolTUI.selectFromTable(available_images,
                                                                 object_type=ExegolImage,
                                                                 allow_none=False)
            except IndexError:
                # No images are available
                if install_mode:
                    # If no image are available in install mode,
                    # either the user does not have internet,
                    # or the image repository has been changed and no docker image is available
                    logger.critical("Exegol can't be installed offline")
                return None
        else:
            try:
                # Find image by name
                selected_image = await DockerUtils().getOfficialImageFromList(tag)
            except ObjectNotFound:
                logger.error(f"Image '{tag}' was not found. If you wanted to build a local image, you can use the 'build' action instead.")
                return None

        if selected_image is not None and type(selected_image) is ExegolImage:
            # Update existing ExegolImage
            if await DockerUtils().downloadImage(selected_image, install_mode):
                sync_result = None
                # Name comparison allow detecting images without version tag
                if not selected_image.isVersionSpecific() and selected_image.hasVersionTag():
                    async with ExegolStatus(f"Synchronizing version tag information. Please wait.", spinner_style="blue"):
                        # Download associated version tag.
                        sync_result = await DockerUtils().downloadVersionTag(selected_image)
                    # Detect if an error have been triggered during the download
                    if type(sync_result) is str:
                        logger.error(f"Error while downloading version tag, {sync_result}")
                        sync_result = None
                # if version tag have been successfully download, returning ExegolImage from docker response
                if sync_result is not None and type(sync_result) is ExegolImage:
                    return sync_result
                # Version-specific images must skip cache to avoid loading latest image
                return await DockerUtils().getInstalledImage(selected_image.getName(), selected_image.getRepository(), skip_cache=selected_image.isVersionSpecific())
        else:
            # Unknown use case
            logger.critical(f"Unknown selected image '{selected_image}'. Exiting.")
        return cast(Optional[ExegolImage], selected_image)

    @classmethod
    async def updateWrapper(cls) -> bool:
        """Update wrapper source code from git"""
        result = await cls.__updateGit(await ExegolModules().getWrapperGit())
        if result:
            await cls.__untagUpdateAvailable()
            logger.empty_line()
            logger.warning("After this wrapper update, remember to update Exegol [bold]requirements[bold]! ([magenta]python3 -m pip install --upgrade -r requirements.txt[/magenta])")
            logger.empty_line()
        return result

    @classmethod
    async def updateImageSource(cls) -> bool:
        """Update image source code from git submodule"""
        return await cls.__updateGit(await ExegolModules().getSourceGit())

    @classmethod
    async def updateResources(cls) -> bool:
        """Update Exegol-resources from git (submodule)"""
        if not UserConfig().enable_exegol_resources:
            logger.info("Skipping disabled Exegol resources.")
            return False
        try:
            if not await ExegolModules().isExegolResourcesReady() and not await ExegolRich.Confirm('Do you want to update exegol resources.', default=True):
                return False
            return await cls.__updateGit(await ExegolModules().getResourcesGit())
        except CancelOperation:
            # Error during installation, skipping operation
            return False

    @staticmethod
    async def __updateGit(gitUtils: GitUtils) -> bool:
        """User procedure to update local git repository"""
        if ParametersManager().offline_mode:
            logger.error("It's not possible to update a repository in offline mode ...")
            return False
        if not gitUtils.isAvailable:
            logger.empty_line()
            return False
        logger.info(f"Updating Exegol [green]{gitUtils.getName()}[/green] {gitUtils.getSubject()}")
        # Check if pending change -> cancel
        if not gitUtils.safeCheck():
            logger.error("Aborting git update.")
            logger.empty_line()
            return False
        current_branch = gitUtils.getCurrentBranch()
        if current_branch is None:
            logger.warning("HEAD is detached. Please checkout to an existing branch.")
            current_branch = "unknown"
        if logger.isEnabledFor(ExeLog.VERBOSE):
            available_branches = gitUtils.listBranch()
            # Ask to checkout only if there is more than one branch available
            if len(available_branches) > 1:
                logger.info(f"Current git branch : {current_branch}")
                # List & Select git branch
                if current_branch == 'unknown' or current_branch not in available_branches:
                    if "main" in available_branches:
                        default_choice = "main"
                    elif "master" in available_branches:
                        default_choice = "master"
                    else:
                        default_choice = None
                else:
                    default_choice = current_branch
                selected_branch = cast(str, await ExegolTUI.selectFromList(gitUtils.listBranch(),
                                                                           subject="a git branch",
                                                                           title="[not italic]:palm_tree: [/not italic][gold3]Branch[gold3]",
                                                                           default=default_choice))
            elif len(available_branches) == 0:
                logger.warning("No branch were detected!")
                selected_branch = None
            else:
                # Automatically select the only branch in case of HEAD detachment
                selected_branch = available_branches[0]
            # Checkout new branch
            if selected_branch is not None and selected_branch != current_branch:
                gitUtils.checkout(selected_branch)
        # git pull
        return await gitUtils.update()

    @classmethod
    async def checkForWrapperUpdate(cls) -> bool:
        """Check if there is an exegol wrapper update available.
        Return true if an update is available."""
        logger.debug(f"Last wrapper update check: {DataCache().get_wrapper_data().metadata.get_last_check_text()}")
        # Skipping update check
        if DataCache().get_wrapper_data().metadata.is_outdated() and not ParametersManager().offline_mode:
            logger.debug("Running update check")
            return await cls.__checkUpdate()
        return False

    @classmethod
    async def __checkUpdate(cls) -> bool:
        """Depending on the current version (dev or latest) the method used to find the latest version is not the same.
        For the stable version, the latest version is fetch from GitHub release.
        In dev mode, git is used to find if there is some update available."""
        isUpToDate = True
        remote_version = ""
        current_version = ConstantConfig.version
        async with ExegolStatus("Checking for wrapper update. Please wait.", spinner_style="blue"):
            if re.search(r'[a-z]', ConstantConfig.version, re.IGNORECASE):
                # Dev version have a letter in the version code and must check updates via git
                logger.debug("Checking update using: dev mode")
                module = await ExegolModules().getWrapperGit(fast_load=True)
                if module.isAvailable:
                    isUpToDate = module.isUpToDate()
                    last_commit = module.get_latest_commit()
                    remote_version = "?" if last_commit is None else str(last_commit)[:8]
                    current_version = str(module.get_current_commit())[:8]
                else:
                    # If Exegol have not been installed from git clone. Auto-check update in this case is only available from mates release
                    logger.verbose("Auto-update checking is not available in the current context")
            else:
                # If there is no letter, it's a stable release, and we can compare faster with the latest git tag
                logger.debug("Checking update using: stable mode")
                try:
                    remote_version = WebRegistryUtils.getLatestWrapperRelease()
                    # On some edge case, remote_version might be None if there is problem
                    if remote_version is None:
                        raise CancelOperation
                    isUpToDate = cls.__compareVersion(remote_version)
                except CancelOperation:
                    # No internet, postpone update check
                    return False

        if not isUpToDate:
            await cls.__tagUpdateAvailable(remote_version, current_version)
        cls.__updateLastCheckTimestamp()
        return not isUpToDate

    @classmethod
    def __updateLastCheckTimestamp(cls) -> None:
        """Update the last_check metadata timestamp with the current date to avoid multiple update checks."""
        DataCache().get_wrapper_data().metadata.update_last_check()
        DataCache().save_updates()

    @classmethod
    def __compareVersion(cls, version: str) -> bool:
        """Compare a remote version with the current one to check if a new release is available."""
        isUpToDate = True
        try:
            for i in range(len(version.split('.'))):
                remote = int(version.split('.')[i])
                local = int(ConstantConfig.version.split('.')[i])
                if remote > local:
                    isUpToDate = False
                    break
        except ValueError:
            logger.warning(f'Unable to parse Exegol version : {version} / {ConstantConfig.version}')
        return isUpToDate

    @classmethod
    async def __get_current_version(cls) -> str:
        """Get the current version of the exegol wrapper. Handle dev version and release stable version depending on the current version."""
        current_version = ConstantConfig.version
        if re.search(r'[a-z]', ConstantConfig.version, re.IGNORECASE) and ConstantConfig.git_source_installation:
            module = await ExegolModules().getWrapperGit(fast_load=True)
            if module.isAvailable:
                current_version = str(module.get_current_commit())[:8]
        return current_version

    @staticmethod
    async def display_current_version() -> str:
        """Get the current version of the exegol wrapper. Handle dev version and release stable version depending on the current version."""
        version_details = ""
        if ConstantConfig.git_source_installation:
            module = await ExegolModules().getWrapperGit(fast_load=True)
            if module.isAvailable:
                current_branch = module.getCurrentBranch()
                commit_version = ""
                if re.search(r'[a-z]', ConstantConfig.version, re.IGNORECASE):
                    commit_version = "-" + str(module.get_current_commit())[:8]
                if current_branch is None:
                    current_branch = "HEAD"
                if current_branch != "master" or commit_version != "":
                    version_details = f" [bright_black]\\[{current_branch}{commit_version}][/bright_black]"
        return f"[blue]v{ConstantConfig.version}[/blue]{version_details}"

    @classmethod
    async def __tagUpdateAvailable(cls, latest_version: str, current_version: Optional[str] = None) -> None:
        """Update the 'update available' cache data."""
        DataCache().get_wrapper_data().last_version = latest_version
        DataCache().get_wrapper_data().current_version = (await cls.__get_current_version()) if current_version is None else current_version

    @classmethod
    async def isUpdateAvailable(cls) -> bool:
        """Check if the cache file is present to announce an available update of the exegol wrapper."""
        current_version = await cls.__get_current_version()
        wrapper_data = DataCache().get_wrapper_data()
        # Check if a latest version exist and if the current version is the same, no external update had occurred
        if wrapper_data.last_version != current_version and wrapper_data.current_version == current_version:
            return True
        else:
            # If the version changed, exegol have been updated externally (via pip for example)
            if wrapper_data.current_version != current_version:
                await cls.__untagUpdateAvailable(current_version)
            return False

    @classmethod
    def display_latest_version(cls) -> str:
        last_version = DataCache().get_wrapper_data().last_version
        if len(last_version) == 8 and '.' not in last_version:
            return f"[bright_black]\\[{last_version}][/bright_black]"
        return f"[blue]v{last_version}[/blue]"

    @classmethod
    async def __untagUpdateAvailable(cls, current_version: Optional[str] = None) -> None:
        """Reset the latest version to the current version"""
        if current_version is None:
            current_version = await cls.__get_current_version()
        DataCache().get_wrapper_data().last_version = current_version
        DataCache().get_wrapper_data().current_version = current_version
        DataCache().save_updates()

    @classmethod
    async def __buildSource(cls, build_name: Optional[str] = None) -> str:
        """build user process :
        Ask user is he want to update the git source (to get new& updated build profiles),
        User choice a build name (if not supplied)
        User select the path to the dockerfiles (only from CLI parameter)
        User select a build profile
        Start docker image building
        Return the name of the built image"""
        # Selecting the default path
        # Don't force update source if using a custom build_path
        if ParametersManager().build_path is None:
            build_path = Path(UserConfig().exegol_images_path)
            # Ask to update git
            try:
                # Install sources and check for update available
                source_git = await ExegolModules().getSourceGit()
                if source_git.isAvailable and not source_git.isUpToDate() and \
                        await ExegolRich.Confirm("Do you want to update image sources (in order to update local build profiles)?", default=True):
                    await cls.updateImageSource()
            except CancelOperation:
                logger.critical("An error prevented exegol from downloading the sources for building a local image.")
            except AssertionError:
                # Catch None git object assertions (from isUpToDate method)
                logger.warning("Git update is [orange3]not available[/orange3]. Skipping.")
        else:
            build_path = Path(ParametersManager().build_path).expanduser().absolute()
            # Check if we have a directory or a file to select the project directory
            if not build_path.is_dir():
                build_path = build_path.parent
            # Check if there is Dockerfile profiles
            if not ((build_path / "Dockerfile").is_file() or len(list(build_path.glob("*.dockerfile"))) > 0):
                logger.critical(f"The directory {build_path} doesn't contain any [green]Dockerfile[/green] or [green]*.dockerfile[/green] build profile.")

        # Choose tag name
        blacklisted_build_name = ["stable", "full", "nightly", "ad", "web", "light", "osint", "free"]
        while build_name is None or build_name in blacklisted_build_name or True in [build_name.startswith(x + '-') for x in blacklisted_build_name]:
            if build_name is not None:
                logger.error("This name is reserved and cannot be used for local build. Please choose another one.")
            build_name = await ExegolRich.Ask("Choose a name for the new local image",
                                              default="local")

        # Choose dockerfiles path
        logger.debug(f"Using {build_path} as path for dockerfiles")

        # Choose dockerfile
        profiles = cls.listBuildProfiles(profiles_path=build_path)
        if len(profiles) == 0:
            logger.critical(f"No build profile found in {build_path}. Check your exegol installation, it seems to be broken...")
        build_profile: Optional[str] = ParametersManager().build_profile
        build_dockerfile: Optional[str] = None
        if build_profile is not None:
            build_dockerfile = profiles.get(build_profile)
            if build_dockerfile is None:
                logger.error(f"Build profile {build_profile} not found.")
        if build_dockerfile is None:
            build_profile, build_dockerfile = cast(Tuple[str, str], await ExegolTUI.selectFromList(profiles,
                                                                                                   subject="a build profile",
                                                                                                   title="[not italic]:dog: [/not italic][gold3]Build profiles[/gold3]"))
        logger.debug(f"Using {build_profile} build profile ({build_dockerfile})")
        # Docker Build
        await DockerUtils().buildImage(tag=build_name, build_profile=build_profile, build_dockerfile=build_dockerfile, dockerfile_path=build_path.as_posix())
        return build_name

    @classmethod
    async def buildAndLoad(cls, load_after_build: bool) -> Optional[ExegolImage]:
        """Build an image and load it"""
        build_name = await cls.__buildSource(ParametersManager().imagetag)
        if load_after_build:
            return await DockerUtils().getInstalledImage(build_name, ConstantConfig.COMMUNITY_IMAGE_NAME)
        return None

    @classmethod
    def listBuildProfiles(cls, profiles_path: Path) -> Dict:
        """List every build profiles available locally
        Return a dict of options {"key = profile name": "value = dockerfile full name"}"""
        # Default stable profile
        profiles = {}
        if (profiles_path / "Dockerfile").is_file():
            profiles["full"] = "Dockerfile"
        # List file *.dockerfile is the build context directory
        logger.debug(f"Loading build profile from {profiles_path}")
        docker_files = list(profiles_path.glob("*.dockerfile"))
        for file in docker_files:
            # Convert every file to the dict format
            filename = file.name
            profile_name = filename.replace(".dockerfile", "")
            profiles[profile_name] = filename
        logger.debug(f"List docker build profiles : {profiles}")
        return profiles

    @classmethod
    async def listGitStatus(cls) -> Sequence[Dict[str, str]]:
        """Get status of every git modules"""
        result = []
        gits = [await ExegolModules().getWrapperGit(fast_load=True),
                await ExegolModules().getSourceGit(fast_load=True, skip_install=True),
                await ExegolModules().getResourcesGit(fast_load=True, skip_install=True)]

        async with ExegolStatus(f"Loading module information", spinner_style="blue") as s:
            for git in gits:
                s.update(status=f"Loading module [green]{git.getName()}[/green] information")
                status = "[bright_black]Unknown[/bright_black]" if ParametersManager().offline_mode else git.getTextStatus()
                branch = git.getCurrentBranch()
                if branch is None:
                    if "not supported" in status or "Not installed" in status:
                        branch = "[bright_black]N/A[/bright_black]"
                    else:
                        branch = "[bright_black][g]? :person_shrugging:[/g][/bright_black]"
                result.append({"name": git.getName().capitalize(),
                               "status": status,
                               "current branch": branch})
        return result



import json
from typing import List

from exegol.model.CacheModels import ImageCacheModel, WrapperCacheModel, CacheDB, ImagesCacheModel
from exegol.utils.DataFileUtils import DataFileUtils
from exegol.utils.ExeLog import logger
from exegol.utils.MetaSingleton import MetaSingleton


class DataCache(DataFileUtils, metaclass=MetaSingleton):
    """This class allows loading cached information defined configurations

    Example of data:
    {
        wrapper: {
            update: {
                metadata: {
                    last_check DATE
                }
                last_version: STR
            }
        }
        images:
            metadata: {
                    last_check: DATE
            }
            data: [
                {
                    name: STR (tag name)
                    last_version: STR (x.y.z|commit_id)
                    type: STR (local|remote)
                }
            ]
    }
    """

    def __init__(self) -> None:
        # Cache data
        self.__cache_data = CacheDB()

        # Config file options
        super().__init__(".datacache", "json")

    def _process_data(self) -> None:
        if len(self._raw_data) >= 2:
            self.__cache_data.load(**self._raw_data)

    def _build_file_content(self) -> str:
        return json.dumps(self.__cache_data, cls=self.ObjectJSONEncoder)

    def save_updates(self) -> None:
        self._create_config_file()

    def get_wrapper_data(self) -> WrapperCacheModel:
        """Get Wrapper information from cache"""
        return self.__cache_data.wrapper

    def get_images_data(self) -> ImagesCacheModel:
        """Get Images information from cache"""
        return self.__cache_data.images

    async def update_image_cache(self, images: List) -> None:
        """Refresh image cache data"""
        logger.debug("Updating image cache data")
        cache_images = []
        for img in images:
            name = img.getName()
            version = img.getLatestVersion()
            if "N/A" in version:
                continue
            remote_id = img.getLatestRemoteId()
            image_type = "local" if img.isLocal() else "remote"
            logger.debug(f"└── {name} (version: {version})\t→ ({image_type}) {remote_id}")
            cache_images.append(
                ImageCacheModel(
                    name,
                    version,
                    remote_id,
                    image_type
                )
            )
        self.__cache_data.images = ImagesCacheModel(cache_images)
        self.save_updates()



import asyncio
import errno
import os
import shutil
import subprocess
import tempfile
from datetime import datetime
from enum import IntFlag, auto as enum_auto
from pathlib import Path
from typing import Optional, Dict, Sequence, Tuple, Union, List

from docker.errors import NotFound, ImageNotFound, APIError
from docker.models.containers import Container

from exegol.config.EnvInfo import EnvInfo
from exegol.console.ExegolPrompt import ExegolRich
from exegol.console.ExegolStatus import ExegolStatus
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import CancelOperation, ObjectNotFound
from exegol.model.ContainerConfig import ContainerConfig
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.SelectableInterface import SelectableInterface
from exegol.utils.ContainerLogStream import ContainerLogStream
from exegol.utils.ExeLog import logger
from exegol.utils.GuiUtils import GuiUtils
from exegol.utils.SessionHandler import SessionHandler
from exegol.utils.imgsync.ImageScriptSync import ImageScriptSync


class ExegolContainer(ExegolContainerTemplate, SelectableInterface):
    """Class of an exegol container already create in docker"""

    class Filters(IntFlag):
        STARTED = enum_auto()
        OUTDATED = enum_auto()

    BACKUP_DIRECTORY = "/workspace/.ExegolUpgradeBackupAuto"

    def __init__(self, docker_container: Container, model: Optional[ExegolContainerTemplate] = None):
        logger.debug(f"Loading container: {docker_container.name}")
        self.__container: Container = docker_container
        self.__id: str = docker_container.id
        self.__xhost_applied = False
        self.__post_start_applied = False
        if model is None:
            image_name = ""
            try:
                # Try to find the attached docker image
                docker_image = docker_container.image
            except ImageNotFound:
                # If it is not found, the user has probably forcibly deleted it manually
                logger.warning(f"Some images were forcibly removed by docker when they were used by existing containers!")
                logger.error(f"The '{docker_container.name}' containers might not work properly anymore and should also be deleted and recreated with a new image.")
                docker_image = None
                image_name = "[red bold]BROKEN[/red bold]"
            # Create Exegol container from an existing docker container
            super().__init__(str(docker_container.name),
                             config=ContainerConfig(container=docker_container),
                             image=ExegolImage(name=image_name, docker_image=docker_image))
            self.image.syncContainerData(docker_container)
            # At this stage, the container image object has an unknown status because no synchronization with a registry has been done.
            # This could be done afterwards (with container.image.autoLoad()) if necessary because it takes time.
            self.__new_container = False
        else:
            # Create Exegol container from a newly created docker container with its object template.
            super().__init__(str(docker_container.name),
                             config=ContainerConfig(container=docker_container),
                             # Rebuild config from docker object to update workspace path
                             image=model.image)
            self.__new_container = True
        self.image.syncStatus()

    def __str__(self) -> str:
        """Default object text formatter, debug only"""
        return f"{self.getRawStatus()} - {super().__str__()}"

    def filter(self, filters: int) -> bool:
        """
        Apply bitwise filter
        :param filters:
        :return:
        """
        match = True
        if match and filters & self.Filters.STARTED:
            match = self.isRunning()
        if match and filters & self.Filters.OUTDATED:
            match = self.image.isLocked()
        return match

    def __getState(self) -> Dict:
        """Technical getter of the container status dict"""
        try:
            self.__container.reload()
        except NotFound:
            return {"State": "Removed"}
        return self.__container.attrs.get("State", {})

    def getRawStatus(self) -> str:
        """Raw text getter of the container status"""
        return self.__getState().get("Status", "unknown")

    def getTextStatus(self) -> str:
        """Formatted text getter of the container status"""
        status = self.getRawStatus().lower()
        if status == "unknown":
            return "Unknown"
        elif status == "exited":
            return "[red]Stopped[/red]"
        elif status == "running":
            return "[green]Running[/green]"
        return f"[orange3]{status}[/orange3]"

    def isNew(self) -> bool:
        """Check if the container has just been created or not"""
        return self.__new_container

    def isRunning(self) -> bool:
        """Check is the container is running. Return bool."""
        return self.getRawStatus() == "running"

    def getFullId(self) -> str:
        """Container's id getter"""
        return self.__id

    def getId(self) -> str:
        """Container's short id getter"""
        return self.__container.short_id

    def getKey(self) -> str:
        """Universal unique key getter (from SelectableInterface)"""
        return self.name

    async def start(self) -> None:
        """Start the docker container"""
        if not self.isRunning():
            self.__preStartChecks()
            logger.info(f"Starting container {self.name}")
            await self.__start_container()
            await self.__postStartSetup()

    async def __start_container(self) -> None:
        """
        This method starts the container and displays startup status updates to the user.
        :return:
        """
        async with ExegolStatus(f"Waiting to start {self.name}", spinner_style="blue") as progress:
            start_date = datetime.now()
            try:
                self.__container.start()
            except APIError as e:
                logger.debug(e)
                logger.critical(f"Docker raised a critical error when starting the container [green]{self.name}[/green], error message is: {e.explanation}")
            if not self.config.legacy_entrypoint:  # TODO improve startup compatibility check
                try:
                    # Try to find log / startup messages. Will time out after 2 seconds if the image don't support status update through container logs.
                    for line in ContainerLogStream(self.__container, start_date=start_date, timeout=2):
                        # Once the last log "READY" is received, the startup sequence is over and the execution can continue
                        if line == "READY":
                            break
                        elif line.startswith('[INFO]'):
                            line = line.replace('[INFO]', '')
                            logger.info(line)
                        elif line.startswith('[VERBOSE]'):
                            line = line.replace('[VERBOSE]', '')
                            logger.verbose(line)
                        elif line.startswith('[ADVANCED]'):
                            line = line.replace('[ADVANCED]', '')
                            logger.advanced(line)
                        elif line.startswith('[DEBUG]'):
                            line = line.replace('[DEBUG]', '')
                            logger.debug(line)
                        elif line.startswith('[WARNING]'):
                            line = line.replace('[WARNING]', '')
                            logger.warning(line)
                        elif line.startswith('[ERROR]'):
                            line = line.replace('[ERROR]', '')
                            logger.error(line)
                        elif line.startswith('[SUCCESS]'):
                            line = line.replace('[SUCCESS]', '')
                            logger.success(line)
                        elif line.startswith('[PROGRESS]'):
                            line = line.replace('[PROGRESS]', '')
                            logger.verbose(line)
                            progress.update(status=f"[blue][Startup][/blue] {line}")
                        else:
                            logger.debug(line)
                    logger.verbose(f"Container started in {(datetime.now() - start_date).seconds} seconds")

                except KeyboardInterrupt:
                    # User can cancel startup logging with ctrl+C
                    logger.warning("Skipping startup status updates (user interruption). Spawning shell now.")

    async def stop(self, timeout: int = 10) -> None:
        """Stop the docker container"""
        if self.isRunning():
            logger.info(f"Stopping container {self.name}")
            async with ExegolStatus(f"Waiting to stop ({timeout}s timeout)", spinner_style="blue"):
                self.__container.stop(timeout=timeout)

    async def spawnShell(self) -> None:
        """Spawn a shell on the docker container"""
        self.__check_start_version()
        logger.info(f"Location of the exegol workspace on the host : {self.config.getHostWorkspacePath()}")
        for device in self.config.getDevices():
            logger.info(f"Shared host device: {device.split(':')[0]}")
        spawn_all_capabilities = (not self.config.getPrivileged() and
                                  "ALL" not in self.config.getCapabilities() and
                                  ParametersManager().capabilities and
                                  "ALL" in ParametersManager().capabilities)
        if not self.__new_container and not spawn_all_capabilities and ParametersManager().capabilities and len(ParametersManager().capabilities) > 0:
            if set(ParametersManager().capabilities).issubset(self.config.getCapabilities()):
                if not self.config.getPrivileged() and "ALL" not in self.config.getCapabilities():
                    logger.warning("Can't set specific capability on existing containers, ignoring. Use [green]--cap ALL[/green] instead if needed.")
            else:
                logger.critical("Can't set specific capability on existing containers. Use [green]--cap ALL[/green] instead if needed.")
        logger.success(f"Opening [blue]{ParametersManager().shell}[/blue] shell in Exegol [green]{self.name}[/green]"
                       f"{' with [orange3]all capabilities[/orange3]' if spawn_all_capabilities or self.config.getPrivileged() or 'ALL' in self.config.getCapabilities() else ''}")
        # In case of multi-user environment, xhost must be set before opening each session to be sure
        await self.__applyX11ACLs()
        # Using system command to attach the shell to the user terminal (stdin / stdout / stderr)
        envs = self.config.getShellEnvs()
        options = ""
        if len(envs) > 0:
            options += f" -e {' -e '.join(envs)}"
        if spawn_all_capabilities:
            options += " --privileged"
        cmd = f"docker exec{options} -ti {self.getFullId()} {self.config.getShellCommand()}"
        logger.debug(f"Opening shell with: {cmd}")
        if EnvInfo.isDockerDesktop() and (EnvInfo.is_windows_shell or EnvInfo.is_mac_shell):
            # Disable "What's next?" Docker Desktop spam exit message
            os.environ['DOCKER_CLI_HINTS'] = "false"
        os.system(cmd)
        # Docker SDK doesn't support (yet) stdin properly
        # result = self.__container.exec_run(ParametersManager().shell, stdout=True, stderr=True, stdin=True, tty=True,
        #                                    environment=self.config.getShellEnvs())
        # logger.debug(result)

    async def exec(self, command: Union[str, Sequence[str]], as_daemon: bool = True, quiet: bool = False, is_tmp: bool = False, show_output: bool = True) -> int:
        """Execute a command / process on the docker container.
        Set as_daemon to not follow the command stream and detach the execution
        Set quiet to disable logs message
        Set is_tmp if the container will automatically be removed after execution"""
        if not self.isRunning():
            await self.start()
        if not quiet:
            logger.info("Executing command on Exegol")
            if logger.getEffectiveLevel() > logger.VERBOSE and not ParametersManager().daemon:
                logger.info("Hint: use verbose mode to see command output (-v).")
        exec_payload, str_cmd = ExegolContainer.formatShellCommand(command, quiet)
        stream = self.__container.exec_run(exec_payload, environment={"CMD": str_cmd, "DISABLE_AUTO_UPDATE": "true"}, detach=as_daemon, stream=not as_daemon and not quiet)
        if as_daemon:
            if not quiet:
                logger.success("Command successfully executed in background")
            return 0
        else:
            try:
                # stream[0] : exit code
                # stream[1] : text stream
                if type(stream.output) is bytes:
                    # When quiet is True, stream is False, so we receive all the logs at the end of the command execution
                    if stream.exit_code is not None and stream.exit_code != 0:
                        logger.debug(f"An error occurred while executing the command: [error {stream.exit_code}] {command}")
                    if len(stream.output) > 0 and show_output:
                        if stream.exit_code is None or stream.exit_code == 0:
                            logger.raw(stream.output.decode("utf-8"))
                        else:
                            logger.error(stream.output.decode("utf-8"))
                else:
                    # When quiet is False, stream is True, so we directly receive a log stream
                    for log in stream[1]:
                        if type(log) is bytes:
                            log = log.decode("utf-8")
                        logger.raw(log)
                if not quiet:
                    logger.success("End of the command")
                return 0 if stream.exit_code is None else stream.exit_code
            except KeyboardInterrupt:
                if not quiet and not is_tmp:
                    logger.info("Detaching process logging")
                    logger.warning("Exiting this command does [red]NOT[/red] stop the process in the container")
        return 0

    @staticmethod
    def formatShellCommand(command: Union[str, Sequence[str]], quiet: bool = False, entrypoint_mode: bool = False) -> Tuple[str, str]:
        """Generic method to format a shell command and support zsh aliases.
        Set quiet to disable any logging here.
        Set entrypoint_mode to start the command with the entrypoint.sh config loader.
        - The first return argument is the payload to execute with every pre-routine for zsh.
        - The second return argument is the command itself in str format."""
        # Using base64 to escape special characters
        str_cmd = command if type(command) is str else ' '.join(command)
        # str_cmd = str_cmd.replace('"', '\\"')  # This fix shoudn' be necessary plus it can alter data like passwd
        if not quiet:
            logger.success(f"Command received: {str_cmd}")
        # ZSH pre-routine: Load zsh aliases and call eval to force aliases interpretation
        cmd = f'autoload -Uz compinit; compinit; source ~/.zshrc; eval "$CMD"'
        if not entrypoint_mode:
            # For direct execution, the full command must be supplied not just the zsh argument
            cmd = f"zsh -c '{cmd}'"
        return cmd, str_cmd

    async def remove(self, container_only: bool = False, backup_history: Optional[List[str]] = None) -> None:
        """Stop and remove the docker container.
        :param container_only: If True, only the container will be removed, not the workspace volume or the network.
        :param backup_history: List for backup containers ID to remove too
        """
        if not container_only:
            await self.__removeVolume()
        await self.stop(timeout=2)
        have_backup = backup_history is not None and len(backup_history) > 0
        backup_text = f" and {len(backup_history)} backup containers" if have_backup and backup_history is not None else ""
        logger.info(f"Removing container {self.name}{backup_text}")
        try:
            self.__container.remove()
            logger.success(f"Container {self.name} successfully removed.")
        except NotFound:
            logger.error(f"The container {self.name} has already been removed (probably created as a temporary container).")
        if not container_only:
            # Must be imported locally to avoid circular importation
            from exegol.utils.DockerUtils import DockerUtils
            if have_backup and backup_history is not None:
                for c_id in backup_history:
                    try:
                        DockerUtils().removeContainerById(c_id)
                    except ObjectNotFound:
                        continue
            nets = self.config.getNetworks()
            for net in nets:
                if net.shouldBeRemoved():
                    logger.debug(f"Network {net.getNetworkName()} will be removed.")
                    DockerUtils().removeNetwork(net.getNetworkName())

    def getExistingBackupContainers(self) -> List[Tuple[str, str]]:
        """
        Get container's backup history and filter existing containers only.
        :return: List of tuple of legacy container. For each entry the first element is the container id and the second is the container name.
        """
        raw_backup_history = self.config.getBackupHistory()
        backup_history = [] if raw_backup_history is None else raw_backup_history.split(',')
        result = []
        if len(backup_history) > 0:
            from exegol.utils.DockerUtils import DockerUtils
            for container_id in backup_history:
                # Check if the previous container still exists
                bak_container = DockerUtils().isContainerExist(container_id)
                if bak_container is not None:
                    result.append((container_id, bak_container[7:]))
        return result

    async def __removeVolume(self) -> None:
        """Remove private workspace volume directory if exist"""
        volume_path = self.config.getPrivateVolumePath()
        # TODO add backup
        if volume_path != '':
            if volume_path.startswith('/wsl/') or volume_path.startswith('\\wsl\\'):
                # Docker volume defines from WSL don't return the real path, they cannot be automatically removed
                # TODO review WSL workspace volume
                logger.warning("Warning: WSL workspace directory cannot be removed automatically.")
                return
            logger.debug(f"Removing volume {volume_path}")
            force_remove = False
            try:
                have_files = False
                for _ in Path(volume_path).iterdir():
                    have_files = True
                    break
            except (PermissionError, OSError) as e:
                if type(e) is OSError:
                    logger.warning(f"Error during workspace files access: {e}")
                    message = f"Exegol cannot access the workspace files [magenta]{volume_path}[/magenta], do you want to delete your container's workspace?"
                else:
                    message = f"Insufficient permission to view workspace files [magenta]{volume_path}[/magenta], do you want to delete your container's workspace?"
                if await ExegolRich.Confirm(message, default=False):
                    # Set have_files to skip directly to rmtree
                    have_files = True
                    # Set force_remove to skip user prompt confirmation again
                    force_remove = True
                else:
                    return
            except FileNotFoundError:
                logger.debug("This workspace has already been removed.")
                return
            try:
                try:
                    if have_files:
                        raise OSError
                    logger.info(f"Removing empty workspace volume")
                    os.rmdir(volume_path)  # This function can only remove an empty directory as failsafe
                except OSError as e:
                    if e.errno is not None and e.errno != errno.ENOTEMPTY:
                        logger.error(f"Receive an error during workspace removal: {e}")
                    # Directory is not empty
                    if (not force_remove and
                            not await ExegolRich.Confirm(f"Workspace [magenta]{volume_path}[/magenta] is not empty, do you want to delete it?",
                                                         default=False)):
                        # User can choose not to delete the workspace on the host
                        return
                    logger.verbose(f"Removing workspace volume")
                    # Try to remove files from the host with user permission (work only without sub-directory)
                    shutil.rmtree(volume_path)
            except PermissionError:
                logger.info(f"Deleting the workspace files from the [green]{self.name}[/green] container as root")
                if not self.isRunning():
                    await self.__start_container()
                # If the host can't remove the container's file and folders, the rm command is exec from the container itself as root
                await self.exec("rm -rf /workspace", as_daemon=False, quiet=True)
                try:
                    shutil.rmtree(volume_path)
                except PermissionError:
                    logger.warning(f"I don't have the rights to remove [magenta]{volume_path}[/magenta] (do it yourself)")
                    return
            logger.success("Private workspace volume removed successfully")
        else:
            logger.warning(f"Externally managed workspaces are [red]NOT[/red] automatically removed by exegol. You can manually remove the directory if it's no longer needed: [magenta]{self.config.getHostWorkspacePath()}[/magenta]")

    def __preStartChecks(self) -> None:
        """Rune pre-start test to avoid some docker errors with a better error message for the user"""
        # Check if VPN file still exists
        vpn = self.config.getVpnConfigPath()
        if vpn and not vpn.exists(follow_symlinks=True):
            logger.error(f"The container [green]{self.name}[/green] VPN configuration no longer exists: [magenta]{vpn}[/magenta]")
            logger.critical("Can't start with a missing config. Please restore the file before starting the container.")
        # Check volumes
        for volume in self.config.getVolumes():
            source = volume.get("Source", "")
            target = volume.get("Target", "")
            # Check if the wayland socket still exists at the same path
            if "wayland" in source and target.startswith("/tmp/wayland") and not Path(source).exists():
                logger.error(f"The container [green]{self.name}[/green] was configured with a wayland socket path that no longer exist: [magenta]{source}[/magenta]")
                logger.critical("Can't start with a missing socket. Please create a new container.")

    async def __postStartSetup(self) -> None:
        """
        Operation to be performed after starting a container
        :return:
        """
        if not self.__post_start_applied:
            self.__post_start_applied = True
            await self.__applyX11ACLs()

    def __check_start_version(self) -> None:
        """
        Check spawn.sh up-to-date status and update the script if needed
        :return:
        """
        # Up-to-date container have the script shared over a volume
        # But legacy container must be checked and the code must be pushed
        if not self.config.isWrapperStartShared():
            # If the spawn.sh if not shared, the version must be compared and the script updated
            current_start = ImageScriptSync.getCurrentStartVersion()
            # Try to parse the spawn version of the container. If an alpha or beta version is in use, the script will always be updated.
            spawn_parsing_cmd = ["/bin/bash", "-c", "egrep '^# Spawn Version:[0-9]+$' /.exegol/spawn.sh 2&>/dev/null || echo ':0' | cut -d ':' -f2"]
            container_version = self.__container.exec_run(spawn_parsing_cmd).output.decode("utf-8").strip()
            if current_start != container_version:
                logger.debug(f"Updating spawn.sh script from version {container_version} to version {current_start}")
                self.__container.put_archive("/", ImageScriptSync.getImageSyncTarData(include_spawn=True))

    async def postCreateSetup(self, is_temporary: bool = False) -> None:
        """
        Operation to be performed after creating a container
        :return:
        """
        # if not a temporary container, apply custom config
        if not is_temporary:
            # Update entrypoint script in the container
            self.__container.put_archive("/", ImageScriptSync.getImageSyncTarData(include_entrypoint=True))
            if self.__container.status.lower() == "created":
                await self.__start_container()
            try:
                await self.__updatePasswd()
            except APIError as e:
                if "is not running" in e.explanation:
                    logger.critical("An unexpected error occurred. Exegol cannot start the container after its creation...")
        # Run post start container actions
        await self.__postStartSetup()

    async def __applyX11ACLs(self) -> None:
        """
        If X11 (GUI) is enabled, allow X11 access on host ACL (if not already allowed) for linux and mac.
        If the host is accessed by SSH, propagate xauth cookie authentication if applicable.
        On Windows host, WSLg X11 don't have xhost ACL.
        :return:
        """
        # On Windows host with WSLg no need to run xhost +local or xauth
        if self.config.isGUIEnable() and not self.__xhost_applied and not EnvInfo.isWindowsHost():
            self.__xhost_applied = True  # Can be applied only once per execution
            if shutil.which("xhost") is None:
                if EnvInfo.is_linux_shell:
                    debug_msg = "Try to install the package [green]xorg-xhost[/green] or maybe you don't have X11 on your host?"
                else:
                    debug_msg = "or you don't have one"
                logger.error(f"The [green]xhost[/green] command is not available on your [bold]host[/bold]. "
                             f"Exegol was unable to allow your container to access your graphical environment ({debug_msg}).")
                return

            logger.debug(f"DISPLAY variable: {GuiUtils.getDisplayEnv()}")
            # Extracts the left part of the display variable to determine if remote access is used
            display_host = GuiUtils.getDisplayEnv().split(':')[0]
            # Left part is empty, local access is used to start Exegol
            if display_host == '' or EnvInfo.isMacHost():
                logger.debug("Connecting to container from local GUI, no X11 forwarding to set up")
                # TODO verify that the display format is the same on macOS, otherwise might not set up xauth and xhost correctly
                if EnvInfo.isMacHost():
                    logger.debug(f"Adding xhost ACL to localhost")
                    # add xquartz inet ACL
                    async with ExegolStatus(f"Starting XQuartz...", spinner_style="blue"):
                        os.system(f"xhost + localhost > /dev/null")
                elif not EnvInfo.isWindowsHost():
                    logger.debug(f"Adding xhost ACL to local:{self.config.getUsername()}")
                    # add linux local ACL
                    os.system(f"xhost +local:{self.config.getUsername()} > /dev/null")
                return

            if shutil.which("xauth") is None:
                if EnvInfo.is_linux_shell:
                    debug_msg = "Try to install the package [green]xorg-xauth[/green] to support X11 forwarding in your current environment?"
                else:
                    debug_msg = "or it might not be supported for now"
                logger.error(f"The [green]xauth[/green] command is not available on your [bold]host[/bold]. "
                             f"Exegol was unable to allow your container to access your graphical environment ({debug_msg}).")
                return

            # If the left part of the display variable is "localhost", x11 socket is exposed only on loopback and remote access is used
            # If the container is not in host mode, it won't be able to reach the loopback interface of the host
            if display_host == "localhost" and not self.config.isNetworkHost():
                logger.warning("X11 forwarding won't work on a bridged container unless you specify \"X11UseLocalhost no\" in your host sshd_config")
                logger.warning("[red]Be aware[/red] changing \"X11UseLocalhost\" value can [red]expose your device[/red], correct firewalling is [red]required[/red]")
                # TODO Add documentation to restrict the exposure of the x11 socket to the docker subnet
                return

            # Extracting the xauth cookie corresponding to the current display to a temporary file and reading it from there (grep cannot be used because display names are not accurate enough)
            _, tmpXauthority = tempfile.mkstemp()
            logger.debug(f"Extracting xauth entries to {tmpXauthority}")
            os.system(f"xauth extract {tmpXauthority} $DISPLAY > /dev/null 2>&1")
            xauthEntry = subprocess.check_output(f"xauth -f {tmpXauthority} list 2>/dev/null", shell=True).decode()
            logger.debug(f"xauthEntry to propagate: {xauthEntry}")

            # Replacing the hostname with localhost to support loopback exposed x11 socket and container in host mode (loopback is the same)
            if display_host == "localhost":
                logger.debug("X11UseLocalhost directive is set to \"yes\" or unspecified, X11 connections can be received only on loopback")
                # Modifing the entry to convert <hostname>/unix:<display_number> to localhost:<display_number>
                xauthEntry = f"localhost:{xauthEntry.split(':')[1]}"
            else:
                # TODO latter implement a check to see if the x11 socket is correctly firewalled and warn the user if it is not
                logger.debug("X11UseLocalhost directive is set to \"no\", X11 connections can be received from anywhere")

            # Check if the host has a xauth entry corresponding to the current display.
            if xauthEntry:
                logger.debug(f"Adding xauth cookie to container: {xauthEntry}")
                await self.exec(f"xauth add {xauthEntry}", as_daemon=False, quiet=True)
                logger.debug(f"Removing {tmpXauthority}")
                os.remove(tmpXauthority)
            else:
                logger.warning(f"No xauth cookie corresponding to the current display was found.")

    async def __updatePasswd(self) -> None:
        """
        If configured, update the password of the user inside the container.
        :return:
        """
        if self.config.getPasswd() is not None:
            logger.debug(f"Updating the {self.config.getUsername()} password inside the container")
            await self.exec(f"echo '{self.config.getUsername()}:{self.config.getPasswd()}' | chpasswd", quiet=True)

    def rename_as_old(self):
        import re
        old_match = re.search(r"(.*-bak)(\d*)$", self.getContainerName())
        i = 0 if not old_match else ((int(old_match.group(2)) if old_match.group(2) else 0) + 1)
        base_name = self.getContainerName() + "-bak" if not old_match else old_match.group(1)
        while True:
            new_name = base_name
            if i > 0:
                new_name += str(i)
            logger.debug(f"Renaming container {self.getContainerName()} as {new_name}")
            try:
                self.__container.rename(new_name)
                logger.success(f"Your previous container [orange3]{self.name}[/orange3] has been renamed to [green]{new_name[7:]}[/green] as a backup. You will need to delete it manually when it is no longer needed.")
                return
            except APIError as e:
                if e.status_code == 409 and "is already in use by" in e.explanation:
                    i += 1
                    logger.debug(f"Container {new_name} already exists, trying again")
                    continue
                else:
                    logger.error(e.explanation)
                    logger.critical(f"An unexpected error occurred from docker. Exegol cannot rename the container [green]{self.name}[/green]. Aborting the operation.")

    async def backup(self, backup_exh: bool) -> None:
        """
        This function automatically backup the exegol workspace and history to the workspace directory.
        :return:
        """
        if not SessionHandler().pro_feature_access():
            logger.critical("Exegol backup is only available for Pro or Enterprise users.")
        async with ExegolStatus(f"Ongoing backup of container [green]{self.name}[/green]", spinner_style="blue"):
            result = await self.exec(f"mkdir {self.BACKUP_DIRECTORY}", quiet=True, as_daemon=False)
            if result != 0:
                logger.error(f"The directory {self.BACKUP_DIRECTORY} already exist in the container [green]{self.name}[/green]. "
                             f"It needs to be removed manually before trying to backup this container again.")
                raise CancelOperation

            backup_tasks = [
                # Backup bash/zsh history
                self.exec(f"tail -n +$(($(grep -n '# -=-=-=-=-=-=-=- YOUR COMMANDS BELOW -=-=-=-=-=-=-=- #' ~/.zsh_history | tail -n1 | cut -d ':' -f1) + 1)) ~/.zsh_history > {self.BACKUP_DIRECTORY}/zsh_history", quiet=True, as_daemon=False),
                self.exec(f"tail -n +$(($(grep -n '# -=-=-=-=-=-=-=- YOUR COMMANDS BELOW -=-=-=-=-=-=-=- #' ~/.bash_history | tail -n1 | cut -d ':' -f1) + 1)) ~/.bash_history > {self.BACKUP_DIRECTORY}/bash_history", quiet=True, as_daemon=False),
                # Backup files
                self.exec(f"cp -a /etc/hosts {self.BACKUP_DIRECTORY}/hosts", quiet=True, as_daemon=False),
                self.exec(f"cp -a /etc/resolv.conf {self.BACKUP_DIRECTORY}/resolv.conf", quiet=True, as_daemon=False),
                self.exec(f"cp -a /etc/proxychains.conf {self.BACKUP_DIRECTORY}/proxychains.conf", quiet=True, as_daemon=False),
                self.exec(f"triliumnext-stop && "
                          f"mkdir {self.BACKUP_DIRECTORY}/triliumnext_data && "
                          f"cp -a /opt/tools/triliumnext/data/document.db* {self.BACKUP_DIRECTORY}/triliumnext_data/ && "
                          f"cp -a /opt/tools/triliumnext/data/session_secret.txt {self.BACKUP_DIRECTORY}/triliumnext_data/ && "
                          f"cp -a /opt/tools/triliumnext/data/sessions {self.BACKUP_DIRECTORY}/triliumnext_data/", quiet=True, as_daemon=False),
                self.exec(f"[ $(sed-comment-line /opt/tools/Exegol-history/profile.sh | sed-empty-line | wc -l) -gt 0 ] || return 0 && cp -a /opt/tools/Exegol-history/profile.sh {self.BACKUP_DIRECTORY}/exh_profile.sh", quiet=True, as_daemon=False),
            ]
            if backup_exh:
                # Backup exegol-history
                backup_tasks.append(
                    self.exec(f"[ -d ~/.exegol_history ] || return 0 && exh export creds --format CSV --file {self.BACKUP_DIRECTORY}/exegol_history_creds.csv && exh export hosts --format CSV --file {self.BACKUP_DIRECTORY}/exegol_history_hosts.csv", quiet=True, as_daemon=False),
                )

            results = await asyncio.gather(*backup_tasks)
            for r in results:
                if r != 0:
                    logger.error(f"An error occurred during backup creation of [green]{self.name}[/green] container. Please check the logs above for more information.")
                    raise CancelOperation

    async def restore(self) -> bool:
        """
        This function automatically restores a backup from the exegol workspace.
        :return:
        """
        if not SessionHandler().pro_feature_access():
            logger.critical("Exegol restore is only available for Pro or Enterprise users.")

        async with ExegolStatus(f"Restoring backup of container [green]{self.name}[/green]", spinner_style="blue"):
            results = await asyncio.gather(
                self.exec(f"cat {self.BACKUP_DIRECTORY}/zsh_history >> ~/.zsh_history", quiet=True, as_daemon=False),
                self.exec(f"cat {self.BACKUP_DIRECTORY}/bash_history >> ~/.bash_history", quiet=True, as_daemon=False),
                self.exec(f"cp -a {self.BACKUP_DIRECTORY}/hosts /etc/hosts", quiet=True, as_daemon=False),
                self.exec(f"cp -a {self.BACKUP_DIRECTORY}/resolv.conf /etc/resolv.conf", quiet=True, as_daemon=False),
                self.exec(f"mv {self.BACKUP_DIRECTORY}/proxychains.conf /etc/proxychains.conf ", quiet=True, as_daemon=False),
                self.exec(f"rm /opt/tools/triliumnext/data/document.db* && "
                          f"mv {self.BACKUP_DIRECTORY}/triliumnext_data/document.db* /opt/tools/triliumnext/data/ && "
                          f"mv {self.BACKUP_DIRECTORY}/triliumnext_data/session_secret.txt /opt/tools/triliumnext/data/session_secret.txt && "
                          f"rm -r /opt/tools/triliumnext/data/sessions && "
                          f"mv {self.BACKUP_DIRECTORY}/triliumnext_data/sessions /opt/tools/triliumnext/data/sessions", quiet=True, as_daemon=False),
                self.exec(f"[ -f {self.BACKUP_DIRECTORY}/exh_profile.sh ] || return 0 && mv {self.BACKUP_DIRECTORY}/exh_profile.sh /opt/tools/Exegol-history/profile.sh", quiet=True, as_daemon=False),
                self.exec(f"[ -f {self.BACKUP_DIRECTORY}/exegol_history_creds.csv ] || return 0 && exh version 2> /dev/null || (echo 'This image is not up-to-date, [green]exegol-history[/green] cannot restore your creds database. Backup can be found in your container here: {self.BACKUP_DIRECTORY}/exegol_history_creds.csv' && return 1) && exh import creds --format CSV --file {self.BACKUP_DIRECTORY}/exegol_history_creds.csv", quiet=True, as_daemon=False),
                self.exec(f"[ -f {self.BACKUP_DIRECTORY}/exegol_history_hosts.csv ] || return 0 && exh version 2> /dev/null || (echo 'This image is not up-to-date, [green]exegol-history[/green] cannot restore your hosts database. Backup can be found in your container here: {self.BACKUP_DIRECTORY}/exegol_history_hosts.csv' && return 1) && exh import hosts --format CSV --file {self.BACKUP_DIRECTORY}/exegol_history_hosts.csv", quiet=True, as_daemon=False),
            )

            for r in results:
                if r != 0:
                    logger.error(f"An error occurred during backup restoration, aborting. Please check the logs above for more information.")
                    logger.error(f"{self.BACKUP_DIRECTORY} backup directory will not be removed automatically, you will have to remove it manually inside the container. ")
                    return False

            await self.exec(f"rm -rf {self.BACKUP_DIRECTORY}", quiet=True, as_daemon=False)
            return True



from enum import Enum
from typing import Optional, Union, List, Tuple


class DockerDrivers(Enum):
    """Enum for Docker driver type"""
    Disabled = 'none'
    Host = 'host'
    Bridge = 'bridge'


class ExegolNetworkMode(Enum):
    """Enum for user display.
    Key is the CLI parameter, Value is the display name"""
    disabled = 'Disabled'
    host = 'Host'
    docker = 'Docker'
    nat = 'NAT'  # need pre-process
    attached = 'External'  # need pre-process


class ExegolNetwork:
    DEFAULT_DOCKER_NETWORK = [d.value for d in DockerDrivers]

    __DEFAULT_NETWORK_DRIVER = DockerDrivers.Bridge

    def __init__(self, net_mode: ExegolNetworkMode = ExegolNetworkMode.host, net_name: Optional[str] = None) -> None:
        self.__net_mode: ExegolNetworkMode = net_mode
        try:
            # Handle Disable and Host drivers
            self.__docker_net_mode = DockerDrivers[self.__net_mode.value]
        except KeyError:
            self.__docker_net_mode = self.__DEFAULT_NETWORK_DRIVER
        self.__net_name = net_name if net_name else self.__docker_net_mode.value

    @classmethod
    def instance_network(cls, mode: Union[ExegolNetworkMode, str], container_name: str) -> "ExegolNetwork":
        if type(mode) is str:
            return cls(net_mode=ExegolNetworkMode.attached, net_name=mode)
        else:
            assert type(mode) is ExegolNetworkMode
            if mode in [ExegolNetworkMode.host, ExegolNetworkMode.docker]:
                return cls(net_mode=mode)
            elif mode == ExegolNetworkMode.nat:
                return cls(net_mode=mode, net_name=container_name)
            elif mode == ExegolNetworkMode.disabled:
                raise ValueError("Network disable cannot be created")
        raise NotImplementedError("This network type is not implemented yet.")

    @classmethod
    def parse_networks(cls, networks: dict, container_name: str) -> List["ExegolNetwork"]:
        results = []
        for network, config in networks.items():
            if network == DockerDrivers.Host.value:
                net_mode = ExegolNetworkMode.host
            elif network == DockerDrivers.Bridge.value:
                net_mode = ExegolNetworkMode.docker
            elif network == container_name:
                net_mode = ExegolNetworkMode.nat
            else:
                net_mode = ExegolNetworkMode.attached
            results.append(cls(net_mode=net_mode, net_name=network))

        return results

    def getNetworkConfig(self) -> Tuple[str, str]:
        return self.__net_name, self.__docker_net_mode.value

    def getNetworkMode(self) -> ExegolNetworkMode:
        return self.__net_mode

    def getNetworkDriver(self) -> DockerDrivers:
        return self.__docker_net_mode

    def getNetworkName(self) -> str:
        return self.__net_name

    def getTextNetworkMode(self) -> str:
        if self.__net_mode is ExegolNetworkMode.attached:
            return self.__net_name
        return self.__net_mode.value

    def shouldBeRemoved(self) -> bool:
        return self.__net_mode == ExegolNetworkMode.nat

    def __repr__(self) -> str:
        repr_str = self.__net_mode.value
        if self.__net_mode in [ExegolNetworkMode.nat, ExegolNetworkMode.attached]:
            repr_str += f" ({self.__net_name} : {self.__docker_net_mode.value})"
        return repr_str



import json
import os
import re
import time
from typing import Any, Optional, Dict, List, Union

import requests
from docker.models.images import Image
from requests import Response
from rich.status import Status

from exegol.config.ConstantConfig import ConstantConfig
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import CancelOperation
from exegol.model.MetaImages import MetaImages
from exegol.utils.ExeLog import logger, ExeLog


class WebRegistryUtils:
    __registry_token: Optional[str] = None

    @classmethod
    def listRemoteDockerhubImages(cls, image: str, status: Status) -> List[MetaImages]:
        """
        List remote dockerhub images available.
        :param image: Name of docker image from Dockerhub
        :param status: Status of Rich Progress TUI
        Return a list of ExegolImage
        """
        logger.debug(f"Fetching remote image tags, digests and sizes from {image}")
        status.update(status=f"Fetching registry information from [green]{image}[/green]")
        remote_results = []
        # Define max number of tags to download from dockerhub (in order to limit download time and discard historical versions)
        page_size = 20
        page_max = 3
        current_page = 1
        url: Optional[str] = f"https://{ConstantConfig.DOCKER_HUB}/v2/repositories/{image}/tags?page=1&page_size={page_size}"
        # Handle multi-page tags from registry
        while url is not None:
            if current_page > page_max:
                logger.debug("Max page limit reached. In non-verbose mode, downloads will stop there.")
                if not logger.isEnabledFor(ExeLog.VERBOSE):
                    break
            current_page += 1
            if logger.isEnabledFor(ExeLog.VERBOSE):
                status.update(status=f"Fetching registry information from [green]{url}[/green]")
            # TODO auth no supported
            docker_repo_response = cls.runJsonRequest(url, "Dockerhub")
            if docker_repo_response is None:
                logger.warning("Skipping online queries.")
                return []
            error_message = docker_repo_response.get("message")
            if error_message:
                logger.error(f"Dockerhub send an error message: {error_message}")
            for dockerhub_data in docker_repo_response.get("results", []):
                # Parse dockerhub data to MetaImage format
                dockerhub_images: List[Dict[str, Optional[Union[str, int]]]] = [a for a in dockerhub_data.get('images', []) if a.get("architecture") != 'unknown']
                images_size: Dict[str, int] = {}  # Arch : size
                for i in dockerhub_images:
                    size = i.get('size', '0')
                    if size is not None:
                        images_size[cls.parseArch(i)] = int(size)
                digest = dockerhub_data.get("digest")
                tag_name = dockerhub_data.get("name", "")
                if digest is None:
                    if len(images_size.keys()) > 0:
                        logger.debug(f"Missing ID for image {tag_name}, manual fetching ! May slow down the process..")
                        digest = cls.getMetaDigestId(tag_name)
                    else:
                        # Single arch image dont need virtual meta_id
                        digest = dockerhub_images[0].get('digest')
                remote_results.append(MetaImages(digest=digest,
                                                 image_name=image,
                                                 tag_name=tag_name,
                                                 images_size=images_size))
            url = docker_repo_response.get("next")  # handle multiple page tags
        # Remove duplication (version specific / latest release)
        return remote_results

    @staticmethod
    def parseArch(docker_image: Union[Dict[str, Optional[Union[str, int]]], Image]) -> str:
        """Parse and format arch in dockerhub style from registry dict struct.
        Return arch in format 'arch/variant'."""
        arch_key = "architecture"
        variant_key = "variant"
        # Support Docker image struct with specific dict key
        if type(docker_image) is Image:
            docker_image = docker_image.attrs
            arch_key = "Architecture"
            variant_key = "Variant"
        arch = str(docker_image.get(arch_key, "amd64"))
        variant = docker_image.get(variant_key)
        if variant:
            arch += f"/{variant}"
        return arch

    @classmethod
    def __getGuestToken(cls, action: str = "pull", service: str = "registry.docker.io") -> Optional[str]:
        """Generate a guest token for Registry service"""
        url = f"https://auth.docker.io/token?scope=repository:{ConstantConfig.COMMUNITY_IMAGE_NAME}:{action}&service={service}"
        response = cls.runJsonRequest(url, service_name="Docker Auth")
        if response is not None:
            return response.get("access_token")
        logger.error("Unable to authenticate to docker as anonymous")
        logger.debug(response)
        # If token cannot be retrieved, operation must be cancelled
        raise CancelOperation

    @classmethod
    def __generateLoginToken(cls) -> None:
        """Generate an auth token that will be used on future API request"""
        # Currently, only support for guest token
        cls.__registry_token = cls.__getGuestToken()

    @classmethod
    def __getRegistryToken(cls) -> str:
        """Registry auth token getter"""
        if cls.__registry_token is None:
            cls.__generateLoginToken()
        assert cls.__registry_token is not None
        return cls.__registry_token

    @classmethod
    def getLatestWrapperRelease(cls) -> str:
        """Fetch from GitHub release the latest Exegol wrapper version"""
        if ParametersManager().offline_mode:
            raise CancelOperation
        url: str = f"https://api.github.com/repos/{ConstantConfig.GITHUB_REPO}/releases/latest"
        github_response = cls.runJsonRequest(url, "Github")
        if github_response is None:
            raise CancelOperation
        latest_tag = github_response.get("tag_name")
        if latest_tag is None:
            logger.warning("Unable to parse the latest Exegol wrapper version from github API.")
            logger.debug(github_response)
        return latest_tag

    @classmethod
    def getMetaDigestId(cls, tag: str) -> Optional[str]:
        """Get Virtual digest id of a specific image tag from docker registry"""
        if ParametersManager().offline_mode:
            return None
        try:
            token = cls.__getRegistryToken()
        except CancelOperation:
            return None
        manifest_headers = {"Accept": "application/vnd.docker.distribution.manifest.list.v2+json", "Authorization": f"Bearer {token}"}
        # Query Docker registry API on manifest endpoint using tag name
        url = f"https://{ConstantConfig.DOCKER_REGISTRY}/v2/{ConstantConfig.COMMUNITY_IMAGE_NAME}/manifests/{tag}"
        response = cls.__runRequest(url, service_name="Docker Registry", headers=manifest_headers, method="HEAD")
        digest_id: Optional[str] = None
        if response is not None:
            digest_id = response.headers.get("docker-content-digest")
            if digest_id is None:
                digest_id = response.headers.get("etag")
        return digest_id

    @classmethod
    def getRemoteVersion(cls, tag: str) -> Optional[str]:
        """Get image version of a specific image tag from docker registry."""
        if ParametersManager().offline_mode:
            return None
        try:
            token = cls.__getRegistryToken()
        except CancelOperation:
            return None
        # In order to access the metadata of the image, the v1 manifest must be use
        manifest_headers = {"Accept": "application/vnd.docker.distribution.manifest.v1+json", "Authorization": f"Bearer {token}"}
        # Query Docker registry API on manifest endpoint using tag name
        url = f"https://{ConstantConfig.DOCKER_REGISTRY}/v2/{ConstantConfig.COMMUNITY_IMAGE_NAME}/manifests/{tag}"
        response = cls.__runRequest(url, service_name="Docker Registry", headers=manifest_headers, method="GET")
        version: Optional[str] = None
        if response is not None and response.status_code == 200:
            data = json.loads(response.content.decode("utf-8"))
            received_media_type = data.get("mediaType")
            if received_media_type == "application/vnd.docker.distribution.manifest.v1+json":
                # Get image version from legacy v1 manifest (faster)
                # Parse metadata of the current image from v1 schema
                metadata = json.loads(data.get("history", [])[0]['v1Compatibility'])
                # Find version label and extract data
                version = metadata.get("config", {}).get("Labels", {}).get("org.exegol.version", "")

            # Convert image list to a specific image
            elif received_media_type == "application/vnd.docker.distribution.manifest.list.v2+json":
                # Get image version from v2 manifest list (slower)
                # Retrieve image digest id from manifest image list
                manifest = data.get("manifests")
                # Get first image manifest
                # Handle application/vnd.docker.distribution.manifest.list.v2+json spec
                if type(manifest) is list and len(manifest) > 0:
                    # Get Image digest
                    first_digest = manifest[0].get("digest")
                    # Retrieve specific image detail from first image digest (architecture not sensitive)
                    manifest_headers["Accept"] = "application/vnd.docker.distribution.manifest.v2+json"
                    url = f"https://{ConstantConfig.DOCKER_REGISTRY}/v2/{ConstantConfig.COMMUNITY_IMAGE_NAME}/manifests/{first_digest}"
                    response = cls.__runRequest(url, service_name="Docker Registry", headers=manifest_headers, method="GET")
                    if response is not None and response.status_code == 200:
                        data = json.loads(response.content.decode("utf-8"))
                        # Update received media type to ba handle later
                        received_media_type = data.get("mediaType")
            # Try to extract version tag from a specific image
            if received_media_type == "application/vnd.docker.distribution.manifest.v2+json":
                # Get image version from v2 manifest (slower)
                # Retrieve config detail from config digest
                config_digest: Optional[str] = data.get("config", {}).get('digest')
                if config_digest is not None:
                    manifest_headers["Accept"] = "application/json"
                    url = f"https://{ConstantConfig.DOCKER_REGISTRY}/v2/{ConstantConfig.COMMUNITY_IMAGE_NAME}/blobs/{config_digest}"
                    response = cls.__runRequest(url, service_name="Docker Registry", headers=manifest_headers, method="GET")
                    if response is not None and response.status_code == 200:
                        data = json.loads(response.content.decode("utf-8"))
                        # Find version label and extract data
                        version = data.get("config", {}).get("Labels", {}).get("org.exegol.version")
            else:
                logger.debug(f"WARNING: Docker API not supported: {received_media_type}")
        return version

    @classmethod
    def runJsonRequest(cls, url: str, service_name: str, headers: Optional[Dict] = None, method: str = "GET", data: Any = None, retry_count: int = 2) -> Any:
        """Fetch a web page from url and parse the result as json."""
        if ParametersManager().offline_mode:
            return None
        data = cls.__runRequest(url, service_name, headers, method, data, retry_count)
        if data is not None and data.status_code == 200:
            data = json.loads(data.content.decode("utf-8"))
        elif data is not None:
            logger.error(f"Error during web request to {service_name} ({data.status_code}) on {url}")
            if data.status_code == 404 and service_name == "Dockerhub":
                logger.info("The registry is not accessible, if it is a private repository make sure you are authenticated prior with docker login.")
            data = None
        return data

    @classmethod
    def __runRequest(cls, url: str, service_name: str, headers: Optional[Dict] = None, method: str = "GET", data: Any = None, retry_count: int = 2) -> Optional[Response]:
        """Fetch a web page from url and catch / log different use cases.
        Url: web page to quest
        Service_name: service display name for logging purpose
        Retry_count: number of retry allowed."""
        # In case of API timeout, allow retrying 1 more time
        for i in range(retry_count):
            try:
                try:
                    proxies = {}
                    http_proxy = os.environ.get('HTTP_PROXY') or os.environ.get('http_proxy')
                    if http_proxy:
                        proxies['http'] = http_proxy
                    https_proxy = os.environ.get('HTTPS_PROXY') or os.environ.get('https_proxy')
                    if https_proxy:
                        proxies['https'] = https_proxy
                    no_proxy = os.environ.get('NO_PROXY') or os.environ.get('no_proxy')
                    if no_proxy:
                        proxies['no_proxy'] = no_proxy
                    logger.debug(f"Fetching information from {url}")
                    #response = requests.request(method=method, url=url, timeout=(10, 20), verify=ParametersManager().verify, headers=headers, data=data, proxies=proxies if len(proxies) > 0 else None)
                    response = requests.request(method=method, url=url, timeout=(10, 20), verify=True, headers=headers, data=data, proxies=proxies if len(proxies) > 0 else None)
                    return response
                except requests.exceptions.HTTPError as e:
                    if e.response is not None:
                        logger.error(f"Response error: {e.response.content.decode('utf-8')}")
                    else:
                        logger.error(f"Response error: {e}")
                except requests.exceptions.SSLError as err:
                    logger.debug(f"Error: {err}")
                    logger.error(f"Connection error: Unable to establish a secured HTTPS connection with {url}")
                except requests.exceptions.ConnectionError as err:
                    logger.debug(f"Error: {err}")
                    error_re = re.search(r"\[Errno [-\d]+]\s?([^']*)('\))+\)*", str(err))
                    error_msg = ""
                    if error_re:
                        error_msg = f" ({error_re.group(1)})"
                    logger.error(f"Connection error: you probably have no internet.{error_msg}")
                    # Switch to offline mode
                    ParametersManager().offline_mode = True
                except requests.exceptions.RequestException as err:
                    logger.error(f"Unknown connection error: {err}")
                return None
            except requests.exceptions.ReadTimeout:
                logger.verbose(f"{service_name} time out, retrying ({i + 1}/{retry_count}) ...")
                time.sleep(1)
        logger.error(f"[green]{service_name}[/green] request has [red]timed out[/red]. Do you have a slow internet connection, or is the remote service slow/down? Retry later.")
        return None
